'''
 hydrocivil: a package for hydrological methods in civil and enviromental engineering.
 For more information see: https://github.com/lucasglasner/hydrocivil

 Author: Lucas Glasner (lgvivanco96@gmail.com)
 Create Time: 2024-09-DD 16:42:58
 Modified by: Lucas Glasner, 
 Modified time: 2024-10-DD 18:49:12
 Dependencies: numpy, pandas, scipy, matplotlib, rioxarray, rasterio, xarray,
 shapely, geopandas, networkx, whitebox_workflows (optional).
'''


# Core functionality
from . import (
    abstractions,
    geomorphology,
    unithydrographs,
    rain,
    watersheds,
)

# Direct class imports
from .global_vars import _has_whitebox
from .watersheds import RiverBasin
from .rain import RainStorm
from .unithydrographs import LumpedUnitHydrograph

# Utilities
from . import (
    misc,
    web
)

if _has_whitebox:
    from . import wb_tools


__version__ = "0.7.0"
__author__ = "Lucas Glasner"
__email__ = "lgvivanco96@gmail.com"
__license__ = "MIT"

from typing import List
__all__: List[str] = [
    'RiverBasin',
    'LumpedUnitHydrograph',
    'RainStorm'
]



'''
 Author: Lucas Glasner (lgvivanco96@gmail.com)
 Create Time: 2024-08-05 11:11:38
 Modified by: Lucas Glasner,
 Modified time: 2024-08-05 11:11:43
 Description: Main watershed classes
 Dependencies:
'''

import numpy as np
import pandas as pd
import xarray as xr

from typing import Tuple
import geopandas as gpd
from shapely.geometry import LineString, Polygon
import warnings
import whitebox_workflows as wbw


def wbRaster2numpy(obj: wbw.Raster) -> np.ndarray:
    """
    This function grabs a whitebox_workflows Raster object and return
    the image data as a numpy array

    Args:
        obj (whitebox_workflows.Raster): A whitebox Raster object

    Returns:
        (numpy.array): data
    """
    rows = int(np.ceil(obj.configs.rows))
    columns = int(np.ceil(obj.configs.columns))
    nodata = obj.configs.nodata

    # Initialize with nodata
    arr = np.full([rows, columns], np.nan)
    r = 0
    for row in range(0, obj.configs.rows):
        values = obj.get_row_data(row)
        c = 0
        for col in range(0, obj.configs.columns):
            value = values[col]
            if value != nodata:
                arr[r, c] = value
            c += 1
        r += 1
    return arr


def wbRaster2xarray(obj: wbw.Raster, exchange_rowcol: bool = False,
                    flip_y: bool = False, flip_x: bool = False
                    ) -> xr.DataArray:
    """
    This function grabs a whitebox_workflows Raster object and returns
    the image data as an xarray DataArray.

    Args:
        obj (whitebox_workflows.Raster): A whitebox Raster object
        exchange_rowcol (bool, optional): Whether to flip rows and columns.
            Defaults to False.
        flip_y (bool, optional): Whether to flip the y-axis. Defaults to False.
        flip_x (bool, optional): Whether to flip the x-axis. Defaults to False.

    Returns:
        xr.DataArray: The raster data as an xarray DataArray.
    """
    xstart, xend = obj.configs.west, obj.configs.east
    ystart, yend = obj.configs.south, obj.configs.north
    if exchange_rowcol:
        x = np.linspace(xstart, xend, obj.configs.rows)
        y = np.linspace(ystart, yend, obj.configs.columns)[::-1]
    else:
        x = np.linspace(xstart, xend, obj.configs.columns)
        y = np.linspace(ystart, yend, obj.configs.rows)[::-1]

    if flip_y:
        y = y[::-1]
    if flip_x:
        x = x[::-1]

    da = xr.DataArray(data=wbRaster2numpy(obj),
                      dims=['y', 'x'],
                      coords={'x': ('x', x, {'units': obj.configs.xy_units}),
                              'y': ('y', y, {'units': obj.configs.xy_units})},
                      attrs={'title': obj.configs.title,
                             '_FillValue': obj.configs.nodata,
                             'wkt_code': obj.configs.coordinate_ref_system_wkt,
                             'epsg_code': obj.configs.epsg_code})

    return da


def wbAttributes2DataFrame(obj: wbw.Vector) -> pd.DataFrame:
    """
    This function grabs a whitebox_workflows vector object and recuperates
    the attribute table as a pandas dataframe.

    Args:
        obj (whitebox_workflows.Vector): A whitebox Vector object

    Returns:
        df (pandas.DataFrame): Vector Attribute Table 
    """
    attrs = obj.attributes.fields
    names = [field.name for field in attrs]

    df = []
    for c in names:
        values = []
        for i in range(obj.num_records):
            val = obj.get_attribute_value(i, c)
            values.append(val)
        values = pd.Series(values, index=range(obj.num_records), name=c)
        df.append(values)

    df = pd.concat(df, axis=1)
    return df


def wbPoint2geopandas(obj: wbw.Vector, crs: str = None) -> gpd.GeoDataFrame:
    """
    This function transform a whitebox_workflows Point layer to a geopandas
    GeoDataFrame.

    Args:
        obj (whitebox_workflows.Vector): A whitebox vector object with points
        crs (str, optional): Cartographic Projection. Defaults to None.

    Returns:
        gdf (geopandas.GeoDataFrame): Point layer as a GeoDataFrame
    """
    xs = []
    ys = []
    for rec in obj:
        x, y = rec.get_xy_data()
        xs.append(x)
        ys.append(y)
    xs, ys = np.array(xs).squeeze(), np.array(ys).squeeze()
    gdf = gpd.points_from_xy(xs, ys)
    gdf = gpd.GeoDataFrame(geometry=gdf, crs=crs)
    gdf_attrs = wbAttributes2DataFrame(obj)
    gdf = pd.concat([gdf_attrs, gdf], axis=1).set_geometry('geometry')
    return gdf


def wbLine2geopandas(obj: wbw.Vector, crs: str = None) -> gpd.GeoDataFrame:
    """
    This function transform a whitebox_workflows Line layer to a geopandas
    GeoDataFrame.

    Args:
        obj (whitebox_workflows.Vector): A whitebox vector object with lines
        crs (str, optional): Cartographic Projection. Defaults to None.

    Returns:
        gdf (geopandas.GeoDataFrame): Lines as a GeoDataFrame object
    """
    xs = []
    ys = []
    for rec in obj:
        parts = rec.parts
        num_parts = rec.num_parts
        part_num = 1  # actually the next part
        x, y = rec.get_xy_data()
        for i in range(len(x)):
            if part_num < num_parts and i == parts[part_num]:
                xs.append(np.nan)  # discontinuity
                ys.append(np.nan)  # discontinuity
                part_num += 1

            xs.append(x[i])
            ys.append(y[i])
        xs.append(np.nan)  # discontinuity
        ys.append(np.nan)  # discontinuity
    xs, ys = np.array(xs).squeeze(), np.array(ys).squeeze()

    breaks = np.where(np.isnan(xs))[0]
    slices = [slice(None, breaks[0])]
    for i in range(len(breaks)-1):
        slices.append(slice(breaks[i]+1, breaks[i+1]))

    lines = []
    for s in slices:
        line = LineString([(x, y) for x, y in zip(xs[s], ys[s])])
        lines.append(line)

    gdf = gpd.GeoDataFrame(geometry=lines, crs=crs)
    gdf_attrs = wbAttributes2DataFrame(obj)
    gdf = pd.concat([gdf_attrs, gdf], axis=1).set_geometry('geometry')

    return gdf


def wbPolygon2geopandas(obj: wbw.Vector, crs: str = None) -> gpd.GeoDataFrame:
    """
    This function transform a whitebox_workflows Polygon layer to a geopandas
    GeoDataFrame.

    Args:
        obj (whitebox_workflows.Vector): A whitebox vector object with polygons
        crs (str, optional): Cartographic Projection. Defaults to None.

    Returns:
        gdf (geopandas.GeoDataFrame): Polygons as a GeoDataFrame object
    """
    xs = []
    ys = []
    for rec in obj:
        parts = rec.parts
        num_parts = rec.num_parts
        part_num = 1  # actually the next part
        x, y = rec.get_xy_data()
        for i in range(len(x)):
            if part_num < num_parts and i == parts[part_num]:
                xs.append(np.nan)  # discontinuity
                ys.append(np.nan)  # discontinuity
                part_num += 1

            xs.append(x[i])
            ys.append(y[i])

        xs.append(np.nan)  # discontinuity
        ys.append(np.nan)  # discontinuity

    xs, ys = np.array(xs).squeeze(), np.array(ys).squeeze()

    breaks = np.where(np.isnan(xs))[0]
    slices = [slice(None, breaks[0])]
    for i in range(len(breaks)-1):
        slices.append(slice(breaks[i]+1, breaks[i+1]))

    poly = []
    for s in slices:
        line = Polygon([(x, y) for x, y in zip(xs[s], ys[s])])
        poly.append(line)

    gdf = gpd.GeoDataFrame(geometry=poly, crs=crs)
    gdf_attrs = wbAttributes2DataFrame(obj)
    gdf = pd.concat([gdf_attrs, gdf], axis=1).set_geometry('geometry')

    return gdf


def wbVector2geopandas(obj: wbw.Vector, crs: str = None) -> gpd.GeoDataFrame:
    """
    This function transform a whitebox_workflows vector layer to a geopandas
    GeoDataFrame.

    Args:
        obj (whitebox_workflows.Vector): A whitebox Vector object
        crs (str, optional): Cartographic Projection. Defaults to None.

    Returns:
        gdf (geopandas.GeoDataFrame): Vector layer as a GeoDataFrame object
    """
    from whitebox_workflows import VectorGeometryType
    obj_type = obj.header.shape_type.base_shape_type()
    if obj_type == VectorGeometryType.Point:
        return wbPoint2geopandas(obj, crs=crs)

    elif obj_type == VectorGeometryType.PolyLine:
        return wbLine2geopandas(obj, crs=crs)

    else:  # Polygon
        return wbPolygon2geopandas(obj, crs=crs)


def xarray2wbRasterConfigs(da: xr.DataArray) -> wbw.RasterConfigs:
    """
    Generate basic RasterConfigs from an xarray DataArray.

    Args:
        da (xr.DataArray): Input xarray DataArray containing raster data.

    Returns:
        wbw.RasterConfigs: Configuration object for creating a new raster.
    """
    configs = wbw.RasterConfigs()
    dtype_dict = {'float32': wbw.RasterDataType.F32,
                  'float64': wbw.RasterDataType.F64,
                  'int8': wbw.RasterDataType.I8,
                  'int16': wbw.RasterDataType.I16,
                  'int32': wbw.RasterDataType.I32,
                  'int64': wbw.RasterDataType.I64,
                  '<U8': wbw.RasterDataType.U8,
                  '<U16': wbw.RasterDataType.U16,
                  '<U32': wbw.RasterDataType.U32,
                  '<U64': wbw.RasterDataType.U64}
    # Raster shape
    nrows, ncols = da.shape[0], da.shape[1]
    configs.rows = nrows
    configs.columns = ncols
    bounds = da.rio.bounds()
    configs.west = bounds[0]
    configs.east = bounds[2]
    configs.south = bounds[1]
    configs.north = bounds[3]

    # Raster resolution
    dx, dy = da.rio.resolution()
    configs.resolution_x = abs(dx)
    configs.resolution_y = abs(dy)

    # Projection
    try:
        configs.epsg_code = da.rio.crs.to_epsg()
        configs.coordinate_ref_system_wkt = da.rio.crs.to_wkt()
    except Exception as e:
        warnings.warn(str(e))

    # No data and dtype
    configs.nodata = da.rio.nodata
    configs.data_type = dtype_dict[str(da.dtype)]

    return configs


def xarray2wbRaster(da: xr.DataArray) -> wbw.Raster:
    """
    Convert an xarray DataArray to a WhiteboxTools Raster.

    Args:
        da (xr.DataArray): Input xarray DataArray containing raster data.

    Returns:
        wbw.Raster: A new raster created from the DataArray.
    """
    array = da.values
    configs = xarray2wbRasterConfigs(da)
    new_raster = wbw.WbEnvironment().new_raster(configs)
    for row in range(configs.rows):
        for col in range(configs.columns):
            new_raster[row, col] = array[row, col]
    return new_raster


def wbDEMpreprocess(dem: xr.DataArray,
                    depressions_method: str = 'fill',
                    return_streams: bool = False,
                    flow_accumulation_threshold: float = 1e6
                    ) -> Tuple[xr.Dataset, gpd.GeoDataFrame]:
    """
    Preprocess a DEM (Digital Elevation Model) using WhiteboxTools to create
    a depressionless DEM, compute flow direction, flow accumulation, and flow
    length. Optionally, extract stream networks.

    Args:
        dem (xr.DataArray): Input DEM as an xarray DataArray.
        depressions_method (str, optional): Method to remove depressions.
            Options are 'fill' or 'breach'. Defaults to 'fill'.
        return_streams (bool, optional): Whether to extract and return stream
            networks. Defaults to False.
        flow_accumulation_threshold (float, optional): Threshold for flow
            accumulation to define streams. Defaults to 1e6.
    Returns
        Tuple[xr.Dataset, gpd.GeoDataFrame]: A tuple containing:
            - xr.Dataset: Dataset with flow direction, flow accumulation,
                and flow length.
            - gpd.GeoDataFrame: GeoDataFrame with stream networks if
                return_streams is True, otherwise an empty GeoDataFrame.
    """
    # Create whitebox enviroment and transform dem to a whitebox object
    wbe = wbw.WbEnvironment()
    w_dem = xarray2wbRaster(dem)

    # Create the depressionless DEM
    if depressions_method == 'breach':
        w_dem_filled = wbe.breach_depressions_least_cost(w_dem)
    elif depressions_method == 'fill':
        w_dem_filled = wbe.fill_depressions(w_dem)

    # Compute flow direction, accumulation and length
    w_d8flowdir = wbe.d8_pointer(w_dem_filled)
    w_d8flowacc = wbe.d8_flow_accum(w_d8flowdir, input_is_pointer=True,
                                    out_type='catchment area')
    w_d8flowlen = wbe.downslope_flowpath_length(w_d8flowdir)

    # Transform whitebox objects to xarray data arrays
    d8_flowdir = wbRaster2xarray(w_d8flowdir).to_dataset(name='flowdir')
    d8_flowacc = wbRaster2xarray(w_d8flowacc).to_dataset(name='flowacc')
    d8_flowlen = wbRaster2xarray(w_d8flowlen).to_dataset(name='flowlen')
    rasters = xr.merge([d8_flowdir, d8_flowacc, d8_flowlen])
    rasters = rasters.reindex({'x': dem.x, 'y': dem.y}, method='nearest')
    rasters = rasters.where(~np.isnan(dem))

    # Compute vector streams if asked and return final results
    if return_streams:

        w_streams_r = wbe.extract_streams(w_d8flowacc,
                                          flow_accumulation_threshold)
        w_streams_v0 = wbe.raster_streams_to_vector(w_streams_r, w_d8flowdir)
        w_streams_v0 = wbe.vector_stream_network_analysis(w_streams_v0, w_dem)
        streams_v0 = wbVector2geopandas(w_streams_v0[0])
        return (rasters, streams_v0)
    else:
        return (rasters, gpd.GeoDataFrame())



'''
 Author: Lucas Glasner (lgvivanco96@gmail.com)
 Create Time: 1969-12-31 21:00:00
 Modified by: Lucas Glasner, 
 Modified time: 2024-03-21 13:33:18
 Description:
 Dependencies:
'''

import os
import requests


def OpenTopo(DATASET: str, ROI: tuple, OUTPUTPATH: str,
             API_KEY: str = 'YOURKEY') -> requests.models.Response:
    """
    This function makes a request to the OpenTopo server and downloads
    DEM data for a small region. 

    Args:
        DATASET (str): Any of this:
            "SRTMGL3" (SRTM GL3 90m)
            "SRTMGL1" (SRTM GL1 30m)
            "SRTMGL1_E" (SRTM GL1 Ellipsoidal 30m)
            "AW3D30" (ALOS World 3D 30m)
            "AW3D30_E" (ALOS World 3D Ellipsoidal, 30m)
            "SRTM15Plus" (Global Bathymetry SRTM15+ V2.1 500m)
            "NASADEM" (NASADEM Global DEM)
            "COP30" (Copernicus Global DSM 30m)
            "COP90" (Copernicus Global DSM 90m)
            "EU_DTM" (DTM 30m)
            "GEDI_L3" (DTM 1000m)
            "GEBCOIceTopo" (Global Bathymetry 500m)
            "GEBCOSubIceTopo" (Global Bathymetry 500m)

        ROI (tuple): Tuple with coordinates (lonmin, lonmax, latmin, latmax)

        OUTPUTPATH (str): Path to the output file e.g "./blabla/dem.tif"

        API_KEY (str, optional): Your OpenTopo API Key.
            Check https://opentopography.org/developers for details
            Defaults to 'YOURKEY'.

    Returns:
        str: server answer to get request. 
    """
    if API_KEY == 'YOURKEY':
        apiurl = 'https://opentopography.org/developers'
        raise RuntimeError(
            f'Please provide the server API key (Check {apiurl} for details)')
    lonmin, lonmax, latmin, latmax = ROI
    params = {
        'demtype': f'{DATASET}',
        'south': f'{latmin}',
        'north': f'{latmax}',
        'west': f'{lonmin}',
        'east': f'{lonmax}',
        'outputFormat': 'GTiff',
        'API_Key': f'{API_KEY}',
    }
    url = 'https://portal.opentopography.org/API/globaldem'
    response = requests.get(url, params=params)
    with open(f'{OUTPUTPATH}', 'wb') as f:
        f.write(response.content)
    return response



'''
 Author: Lucas Glasner (lgvivanco96@gmail.com)
 Create Time: 1969-12-31 21:00:00
 Modified by: Lucas Glasner,
 Modified time: 2024-05-06 16:40:29
 Description:
 Dependencies:
'''

import os
import warnings
import matplotlib.axes
import matplotlib.figure
import numpy as np
import pandas as pd
import scipy.signal as sg
import geopandas as gpd
import matplotlib

from math import gamma
from typing import Union, Any, Tuple, Type
from numpy.typing import ArrayLike
from scipy.interpolate import interp1d
from scipy.special import gamma

from .geomorphology import concentration_time
from .global_vars import CHILE_UH_LINSLEYPARAMS, CHILE_UH_GRAYPARAMS


# ----------------------------- UNIT HYDROGRAPHS ----------------------------- #

def SUH_Clark(area: float, tc: float, tstep: float,
              R: float = None, X: float = None,
              timearea: pd.Series = None, tolerance: float = 1e-5,
              interp_kwargs: dict = {'kind': 'quadratic'}
              ) -> Tuple[pd.Series, pd.Series]:
    """
    Clark (1945) instantaneous unit hydrograph model. 

    The instantaneous unit hydrograph refers to the catchment response of a
    homogeneous 1 mm storm distributed and applied over an instantaneous
    duration. In this regard the Clark's model indicates that the basin
    response can be computed from a time-area relationship, the concentration
    time and a storage coefficient R. The model assumes that the basin behaves
    like a linear reservoir following the equation:

        dS/dt = I(t) - Q(t)
        S(t) = R * Q(t)

    If a 1mm pulse is homogeneously and instantaneously distributed on the
    basin area the input runoff of Clark linear model will be exactly equal
    to the time-area curve. So, solving the previous equation with a finite
    difference approach leads to: 

        c = (dt / (R + 0.5 * dt))
        Q(t)= c * I(t) + (1 - c) * Q(t-1) 

    which can be solved iteratively from a given time-area curve I(t), 
    a timestep dt and the R storage parameter.

    The time-area curve of the basin is a complex function of the basin 
    geomorphology and hydraulic properties, and should go between 0 in the 
    basin outlet to the concentration time on the farthest hydrological point. 

    The R parameter is a function of the basin properties, however is 
    commonly estimated through empirical formulas. The most common is to 
    compute it from the concentration time with something like:

        X = R / (tc + R)
        R = X * t_c / (1 - X)

    where X is a coefficient determined through regional analyses. From HEC-HMS
    docs: "Smaller values of X result in short, steeply rising unit hydrographs
    and may be representative of urban watersheds. Larger values of X result
    in broad, slowly rising unit hydrographs and may be representative of flat,
    swampy watersheds."

    References:
        Clark, C. O. (1945). Storage and the unit hydrograph. Transactions of
        the American Society of Civil Engineers, 110(1), 1419-1446.

    Args:
        area (float): Basin area (km2)
        tc (float): Basin concentration time (hours)
        R (float): Linear reservoir storage coefficient (hours). Usually an 
            empirical function of the concentration time.
        X (float): Fractional coefficient used to derive R from tc with the
            formula X = R / (tc + R).
        tstep (float): Unit hydrograph discretization time step in hours.
        timearea (pandas.Series, optional): A user defined dimensionless
            time-area curve with dimensionless time (t/tc) in the index and 
            dimensionless area in values (A/A_basin). If None will use the 
            HEC-HMS default time-area curve. Defaults to None.
        tolerance (float, optional): Tolerance for solving the linear
            reservoir ODE. Defaults to 1e-5.
        interp_kwargs (dict, optional): args passed to
            scipy.interpolation.interp1d function. 
            Defaults to {'kind':'quadratic'}.

    Returns:
        _type_: _description_
    """
    text = "SUH_Clark() missing 1 required positional argument: 'R' or 'X'"
    if type(R) == type(None):
        if type(X) != type(None):
            R = X * tc / (1 - X)
        else:
            raise TypeError(text)

    c = 2*tstep/(2*R+tstep)
    if type(timearea) == type(None):
        t_shape = np.arange(0, 1+0.1, 0.1)
        At_shape = np.full(t_shape.shape, np.nan)
        At_shape[t_shape <= 1/2] = 1.414*(t_shape[t_shape <= 1/2])**1.5
        At_shape[t_shape > 1/2] = 1-1.414*(1-t_shape[t_shape > 1/2])**1.5
        timearea = pd.Series(At_shape, index=t_shape)
    timearea.index = timearea.index*tc

    # Interpolate to new time resolution
    ntime = np.arange(timearea.index[0], timearea.index[-1]+tstep, tstep)
    f = interp1d(timearea.index, timearea.values, fill_value='extrapolate',
                 **interp_kwargs)
    timearea = pd.Series(f(ntime), index=ntime)
    timearea = timearea.diff().fillna(0)
    q_in = timearea*1e3/3600

    # Main loop
    uh, time = [0], [0]
    i = 0
    while True:
        i = i+1
        if i < len(q_in):
            qout = q_in.iloc[i]*c+uh[i-1]*(1-c)
        else:
            qout = uh[i-1]*(1-c)
        uh.append(qout)
        time.append(tstep*i)
        if uh[i] < tolerance:
            break
    uh = pd.Series(uh, index=time)
    uh = (uh.shift(1)+uh)/2
    uh = uh.where(uh > 0).fillna(0)
    uh = uh*area

    # Ensure that the unit hydrograph acummulates a volume of 1mm
    volume = np.trapz(uh, uh.index*3600)/1e6/area*1e3
    uh = uh/volume
    params = (uh.max(), uh.idxmax(), tc, R)
    params = pd.Series(params, index=['qpeak', 'tpeak', 'tc', 'R'])
    uh.name, params.name = 'Clark_m3 s-1 mm-1', 'Params_Clark'
    return uh, params


def SUH_SCS(area: float, tc: float, tstep: float, prf: float = 484,
            threshold: float = 1e-10,
            interp_kwargs: dict = {'kind': 'quadratic'}
            ) -> Tuple[pd.Series, pd.Series]:
    """
    U.S.A Soil Conservation Service (SCS) synthetic unit hydrograph.

    The SCS model for unit hydrograph calculation is based on a triangular unit
    hydrograph and a dimensionless unit hydrograph. The model assumes that the
    unit hydrograph can be characterized as a function of the peak flow 'q_p'
    and peak time 't_p'. The equations are:

    t_p = 0.6 * t_c + dt / 2
    q_p = prf * area / t_p

    where t_c is the basin concentration time, dt the reference rain duration
    and timestep and prf the peak rate factor. The dimensionless unit
    hydrograph comes from solving the so called 'gamma equation', which is:

    q / q_p = exp(m) * ((t / t_p) ^ m) * exp(-m * (t / t_p))

    where 'm' is a shape parameter related to the prf by the equation:

    prf = C * m**(m + 1) / exp(m) / Γ(m+1)

    where Γ is the gamma function and C a constant equal to 645.33 for
    imperial units, and 0.2787 for SI units. The peak rate factor
    for the traditional SCS unit hydrograph is equal to PRF = 484 (imperial
    units) or PRF = 0.208 (SI units). However recent literature (?? HEC-HMS ??)
    says it can vary with basin geomorphological parameters for a range of at
    least 100 - 600 (imperial units).

    References:
        Bhunya, P. K., Panda, S. N., & Goel, M. K. (2011).
        Synthetic unit hydrograph methods: a critical review.
        The Open Hydrology Journal, 5(1).

        Natural Resources Conservation Service. Chapter 16 - Hydrographs.
        National Engineering Handbook Part 630 - Hydrology. United States
        Department of Agriculture (USDA), 2007

        Chow Ven, T., Te, C. V., RC, M. D., & Mays, L. W. (1988).
        Applied hydrology. McGraw-Hill Book Company.

    Args:
        area (float): Basin area (km2)
        tc (float): Basin concentration time (hours)
        tstep (float): Unit hydrograph discretization time step in hours.
        prf (float): peak rate factor (imperial units). Defaults to 484. 
        threshold (float): Iteration error for the m parameter calculation 
            based on the given peak rate factor (prf). Defaults to 1e-10. 
        interp_kwargs (dict, optional): args passed to
            scipy.interpolation.interp1d function. 
            Defaults to {'kind':'quadratic'}.

    Returns:
        uh, (qp, tp, tb, tstep) (tuple):
            (unit hydrograph),
            (Peak runoff (L/s/km2/mm), peak time (hours), base time (hours),
            time step (hours)))

    """
    def _iterfunc(target_prf: float, m: float) -> float:
        """
        Args:
            target_prf (float): target peak rate factor
            m (float): input m parameter

        Returns:
            (float): updated m parameter
        """
        C = 645.33
        a = target_prf * np.exp(m)
        b = gamma(m+1) / C
        return (a*b)**(1/(m+1))

    def _solve4m(target_prf: float, threshold: float) -> float:
        """
        This function computes m for a given prf.

        Args:
            target_prf (float): peak rate factor (imperial units)
            threshold (float): Iteration threshold.
        """
        m0 = 0
        while True:
            m = _iterfunc(target_prf, m0)
            if abs(m-m0) < threshold:
                return m
            else:
                m0 = m

    # Unit hydrograph shape
    m = _solve4m(prf, threshold=threshold)
    t_shape = [np.arange(0, 2+0.1, 0.1),
               np.arange(2.2, 4+0.2, 0.2),
               np.arange(4.5, 20+0.5, 0.5)]
    t_shape = np.hstack(t_shape)
    q_shape = np.exp(m)*t_shape**m*np.exp(-m*t_shape)

    # Unit hydrograph paremeters
    tp = tc*0.6+tstep/2
    tb = 2.67*tp
    qp = prf*area/tp*(0.0283/2.58/25.4)  # International units
    uh = pd.Series(qp*q_shape, index=t_shape*tp)

    # Interpolate to new time resolution
    ntime = np.arange(uh.index[0], uh.index[-1]+tstep, tstep)
    f = interp1d(uh.index, uh.values,
                 fill_value='extrapolate', **interp_kwargs)
    uh = f(ntime)
    uh = pd.Series(uh, index=ntime)
    uh = uh.where(uh > 0).fillna(0)

    # Ensure that the unit hydrograph acummulates a volume of 1mm
    volume = np.trapz(uh, uh.index*3600)/1e6/area*1e3
    uh = uh/volume
    params = (qp, tp, tb, tstep, prf)
    params = pd.Series(
        params, index=['prf', 'qpeak', 'tpeak', 'tbase', 'tstep'])
    uh.name, params.name = 'SCS_m3 s-1 mm-1', 'Params_SCS'
    return uh, params


def tstep_correction(tstep: float, tp: float) -> float:
    """
    This functions checks if the selected timestep can be used
    as the unit hydrograph time resolution (unitary time) for Snyder
    or Linsley methods.

    Args:
        tstep (float): Desired time step in hours
        tp (float): Raw unit hydrograph peak time (tu) in hours.

    Raises:
        RuntimeError: If selected tstep exceeds a 10% change of
            unit hydrograph storm duration/peak time.

    Returns:
        float: Fixed timestep and peak time in hours
    """
    tu = tp/5.5
    if tstep > tu*0.5:
        warnings.warn(f'tstep exceeds tu/2, changing tstep to tu = tp/5.5')
        return tu, tp
    if (tstep < tu+0.1) and (tstep > tu-0.1):
        return tstep, tp
    else:
        tp = tp+0.25*(tstep-tu)
        return tstep, tp


def SUH_Gray(area: float, mriverlen: float, meanslope: float,
             a: float, b: float,
             interp_kwargs: dict = {'kind': 'quadratic'}
             ) -> Tuple[pd.Series, pd.Series]:
    """
    Gray's' method assumes a SUH that follows the gamma function, which
    reflects the theoretical result of a basin made of an infinite series
    of lineal reservoirs.

    The model suggests that the dimensionless unit hydrograph can be obtained
    by the equation:

    q_shape = 25 * y^(y+1) * exp(-y * t_shape) * (t_shape)^y / gamma(y + 1)

    where q_shape = q / q_peak and t_shape = t / t_p are the dimensionless
    coordinates of the unit hydrograph. y is the model parameter and gamma
    is the gamma function.

    Based on some regression analysis Gray's method says that the y parameter
    is related to the peak time t_p by the equation:

        t_p / y = (2.676 / t_p + 0.0139)^(-1)
        y = 2.676 + 0.0139 * t_p

    where (t_p / y) would be a measure of the storage property of the watershed,
    or the travel time required for water to pass through a given reach.

    (t_p / y) is naturally dependant of the basin geomorphology so the method
    suggests a relationship of the kind:

        t_p / y = a * ( L / sqrt(S))^b

    where a and b are empirical parameters possibly dependant of geomorphology
    and land cover. Note that this equation in relation with the previous one
    allows to compute the y parameter frmo t_p, so the only real parameters
    of the model are the empirical a and b of the geomorphology equation.


    References:
        Bhunya, P. K., Panda, S. N., & Goel, M. K. (2011).
        Synthetic unit hydrograph methods: a critical review.
        The Open Hydrology Journal, 5(1).

        Bras, R. L. (1990). Hydrology: an introduction to hydrologic science.

    Args:
        area (float): Basin area (km2)
        mriverlen (float): Main channel length (km)
        meanslope (float): Basin mean slope (m/m)
        a, b (float): Gray's model parameters.
        tstep (float): Unit hydrograph target unitary time (tu) in hours.
        interp_kwargs (dict, optional): args passed to
            scipy.interpolation.interp1d function. 
            Defaults to {'kind':'quadratic'}.

    """
    alpha, beta = 2.676, 0.0139
    y = alpha/(1-beta*a*((mriverlen / np.sqrt(meanslope))**b))
    tp = (y-alpha)/beta  # minutes
    tp = tp/60  # hours

    tstep = tp/5.5

    t_shape = np.arange(0, 10+0.1, 0.1)
    q_shape = 25*y**(y+1)*np.exp(-y*t_shape)*(t_shape)**(y)/gamma(y+1)

    uh = pd.Series(q_shape, index=t_shape*tp)
    uh = uh*area/360/(0.25*tp)

    # Interpolate to new time resolution
    ntime = np.arange(uh.index[0], uh.index[-1]+tstep, tstep)
    f = interp1d(uh.index, uh.values, fill_value='extrapolate',
                 **interp_kwargs)
    uh = f(ntime)
    uh = pd.Series(uh, index=ntime)
    uh = uh.where(uh > 0).fillna(0)

    # Ensure that the unit hydrograph acummulates a volume of 1mm
    volume = np.trapz(uh, uh.index*3600)/1e6/area*1e3
    uh = uh/volume
    params = (uh.max(), tp, tstep, y, a, b)
    params = pd.Series(params, index=['qpeak', 'tpeak', 'tstep',
                                      'y', 'a', 'b'])
    uh.name, params.name = 'Gray_m3 s-1 mm-1', 'Params_Gray'
    return uh, params


def SUH_Linsley(area: float, mriverlen: float, out2centroidlen: float,
                meanslope: float,
                C_t: float, n_t: float,
                C_p: float, n_p: float,
                C_b: float, n_b: float,
                interp_kwargs: dict = {'kind': 'quadratic'}
                ) -> Tuple[pd.Series, pd.Series]:
    """
    Linsley unit hydrograph (UH) is a similar formulation of the well known
    Snyder UH. The difference is that Linsley's UH uses a different formula
    for the peak flow like this:

        tp = Ct * (L * Lg / sqrt(S)) ^ nt
        qp = Cp * tp ^ np
        tb = Cb * tp ^ nb

    Where L, Lg and S are basin main channel length, distance between the basin
    outlet and centroid, and basin mean slope.

    Ct, nt are shape parameters for the peak time,
    Cp, np are shape parameters for the peak flow,
    Cb, nb are shape parameters for the base flow

    *** Shape paremeters probably depend on hydrological soil properties,
    land cover use and geomorphology.

    Just like the Snyder UH, the Linsley UH is defined for a rain duration of

        tu = tp / 5.5

    References:
        Manual de calculo de crecidas y caudales minimos en cuencas sin
        informacion fluviometrica. Republica de Chile, Ministerio de Obras
        Publicas (MOP), Dirección General de Aguas (DGA) (1995).

        Metodo para la determinación de los hidrogramas sintéticos en Chile,
        Arteaga F., Benitez A., División de Estudios Hidrológicos,
        Empresa Nacional de Electricidad S.A (1985).

    Args:
        area (float): Basin area (km2)
        mriverlen (float): Main channel length (km)
        out2centroidlen (float): Distance from basin outlet to centroid (km)
        meanslope (float): Basin mean slope (m/m)
        C_t, n_t, C_p, n_p, C_b, n_b (float): Linsley model parameters.
        interp_kwargs (dict, optional): args passed to
            scipy.interpolation.interp1d function. 
            Defaults to {'kind':'quadratic'}.


    Returns:
        uh, (qpR, tpR, tbR, tuR) (tuple):
            (unit hydrograph),
            (Peak runoff (m3/s/km2/mm), peak time (hours), time step (hours)))
    """

    tp = C_t * (mriverlen*out2centroidlen / np.sqrt(meanslope))**n_t

    # Adjust storm duration to the UH timestep
    tpR = tp
    tstep = tp/5.5
    qpR = C_p*tpR**(n_p)
    tbR = C_b*tpR**(n_b)

    # Unit hydrograph shape
    t_shape = np.array([0, 0.3, 0.5, 0.6, 0.75, 1, 1.3, 1.5, 1.8, 2.3, 2.7, 3])
    q_shape = np.array([0, 0.2, 0.4, 0.6, 0.80, 1, 0.8, 0.6, 0.4, 0.2, 0.1, 0])

    # Compute unit hydrograph of duration equals tu
    uh = np.array(q_shape)*qpR
    uh = pd.Series(uh, index=np.array(t_shape)*tpR)

    # Interpolate to new time resolution
    ntime = np.arange(uh.index[0], uh.index[-1] + tstep, tstep)
    f = interp1d(uh.index, uh.values, fill_value='extrapolate',
                 **interp_kwargs)
    uh = f(ntime)
    uh = pd.Series(uh, index=ntime)
    uh = uh.where(uh > 0).fillna(0)

    # Ensure that the unit hydrograph acummulates a volume of 1mm
    volume = np.trapz(uh, uh.index*3600)/1e6  # mm
    uh = uh/volume
    params = (qpR/volume*area/1e3, tpR, tbR,
              tstep, C_t, n_t, C_p, n_p, C_b, n_b)
    params = pd.Series(params, index=['qpeak', 'tpeak', 'tbase', 'tstep',
                                      'C_t', 'n_t', 'C_p', 'n_p',
                                      'C_b', 'n_b'])
    uh.name, params.name = 'Linsley_m3 s-1 mm-1', 'Params_Linsley'
    return uh*area/1e3, params

# ------------------------------- MAIN CLASSES ------------------------------- #


class LumpedUnitHydrograph(object):
    """
    Synthetic Unit Hydrograph class used for building unit hydrograph of
    river basins as a function of geomorphometric and land use properties.
    The class can be used to build the SCS, Linsley or Gray's unit hydrograph
    of any watershed or the pre-calibrated Chilean unit hydrographs of
    DGA/Arteaga&Benitez (Linsley) and Grey's both present in the national flood
    manual.

    Examples:
        + Compute SCS UH for a storm duration of 1 hour.
        + Show params and the related S-Curve
        -> suh = LumpedUnitHydrograph(method='SCS', geoparams=geoparams)
        -> suh.compute(timestep=1)
        -> suh.UnitHydroParams
        -> suh.SHydrograph

        + Compute convolution of unit hydrograph and a series of rainfall
        -> suh.convolve(rainfall)

        + Change duration of unit hydrograph to 30 minutes and compute
        + convolution with a new series of 30 minute rainfall pulses
        -> suh.update_duration(30/60).convolve(rainfall2)
    """

    def __init__(self, method: str,
                 geoparams: Union[dict, pd.Series, pd.DataFrame]) -> None:
        """
        Synthetic unit hydrograph (SUH) constructor.

        Args:
            method (str): Type of synthetic unit hydrograph to use.
                Options: 'SCS', Linsley, ...
            geoparams (__type__): Input geomorphologic and land use parameters.
        """
        self.method = method
        self.geoparams = geoparams
        self.kwargs = {}
        self.timestep = None
        self.UnitHydro = None
        self.SUnitHydro = None
        self.params = None

    def __repr__(self) -> str:
        """
        What to show when invoking a SynthUnitHydro object
        Returns:
            str: Some metadata
        """
        text = f'SynthUnitHydro(method=\'{self.method}\', '
        text = text+f'timestep={self.timestep}'
        if len(self.kwargs.items()) != 0:
            text = text+', '
            text = text+', '.join([f'{x}={y}' for x, y in self.kwargs.items()])
        text = text+')'
        return text

    def _Clark(self, timestep: float, tc_formula: str = 'SCS', **kwargs: Any
               ) -> Tuple[pd.Series, pd.Series]:
        """
        Compute the unit hydrograph following Clark's model.

        Args:
            timestep (float): timestep in hours. Note that this is the same
                as the unit hydrograph excess rain duration.
            tc_formula (str, optional): Empirical formula used for computing
                the time of concentration. Options: 'California', 'Giandotti',
                'Kirpich', 'SCS', 'Spain'. Defaults to 'SCS'.

        Returns:
            (tuple): tuple with the unit hydrograph time series and the
                respective table of parameters
        """
        area = self.geoparams['area']
        if 'tc' in kwargs.keys():
            uh, uh_params = SUH_Clark(tstep=timestep, area=area, **kwargs)
        else:
            tc = concentration_time(method=tc_formula, **self.geoparams)/60
            uh, uh_params = SUH_Clark(tstep=timestep, area=area, tc=tc,
                                      **kwargs)
        return uh, uh_params

    def _SCS(self, timestep: float, tc_formula: str = 'SCS', **kwargs: Any
             ) -> Tuple[pd.Series, pd.Series]:
        """
        Compute the unit hydrograph following the SCS model.

        Args:
            timestep (float): timestep in hours. Note that this is the same
                as the unit hydrograph excess rain duration.
            tc_formula (str, optional): Empirical formula used for computing
                the time of concentration. Options: 'California', 'Giandotti',
                'Kirpich', 'SCS', 'Spain'. Defaults to 'SCS'.

        Returns:
            (tuple): tuple with the unit hydrograph time series and the
                respective table of parameters
        """
        area = self.geoparams['area']
        if 'tc' in kwargs.keys():
            uh, uh_params = SUH_SCS(tstep=timestep, area=area, **kwargs)
        else:
            tc = concentration_time(method=tc_formula, **self.geoparams)/60
            uh, uh_params = SUH_SCS(tstep=timestep, area=area, tc=tc,
                                    **kwargs)
        return uh, uh_params

    def _Linsley(self, DGAChileParams: bool = False, DGAChileZone: bool = None,
                 **kwargs: Any) -> Tuple[pd.Series, pd.Series]:
        """
        Compute the unit hydrograph following the Linsley model.
        Args:
            DGAChileParams (bool, optional): Wheter to use pre-calibrated
                parameters for Chilean basins. Defaults to False.
            DGAChileZone (str, optional): Zone of the pre-calibrated parameters 
                for Chilean basins following Arteaga & Benitez and DGA methods.
                Options: 'I', 'II', 'III', 'IV'. If DGAChileZone == None the 
                function will ask you to give the Linsley model parameters.
                Defaults to None.

        Returns:
            (tuple): tuple with the unit hydrograph time series and the
                respective table of parameters
        """
        geoparams = ['area', 'mriverlen', 'out2centroidlen', 'meanslope']
        if DGAChileParams:
            coefs = CHILE_UH_LINSLEYPARAMS[DGAChileZone]
            uh, uh_params = SUH_Linsley(**self.geoparams[geoparams], **coefs,
                                        **kwargs)
        else:
            uh, uh_params = SUH_Linsley(**self.geoparams[geoparams], **kwargs)
        return uh, uh_params

    def _Gray(self, DGAChileParams: bool = False, **kwargs: Any
              ) -> Tuple[pd.Series, pd.Series]:
        """
        Compute the unit hydrograph following the Gray model.
        Args:
            DGAChileParams (bool, optional): Wheter to use pre-calibrated
                parameters for Chilean basins. Defaults to False.

        Returns:
            (tuple): tuple with the unit hydrograph time series and the
                respective table of parameters
        """
        geoparams = ['area', 'mriverlen', 'meanslope']
        if DGAChileParams:
            a, b = CHILE_UH_GRAYPARAMS['a'], CHILE_UH_GRAYPARAMS['b']
            uh, uh_params = SUH_Gray(**self.geoparams[geoparams], a=a, b=b,
                                     **kwargs)
        else:
            uh, uh_params = SUH_Gray(**self.geoparams[geoparams], **kwargs)
        return uh, uh_params

    def get_SHydrograph(self) -> pd.Series:
        """
        This function computes the S-Curve or S-Hydrograph which is independent
        of the storm duration and can be used for computing the UH of a
        different excess-rain duration.

        Returns:
            (pandas.Series): S-Unit Hydrograph. 
        """
        uh = self.UnitHydro
        sums = [uh.shift(i) for i in range(len(uh)+1)]
        S_uh = pd.concat(sums, axis=1).sum(axis=1)
        return S_uh

    def update_duration(self, duration: float,
                        interp_kwargs: dict = {'kind': 'quadratic'}
                        ) -> Type['LumpedUnitHydrograph']:
        """
        This function uses de S-Curve to update the unit hydrograph duration
        and the respective parameters.

        Args:
            duration (float): New storm duration (equal to time resolution)
            interp_kwargs (dict, optional): args passed to
                scipy.interpolation.interp1d function. 
                Defaults to {'kind':'quadratic'}.


        Returns:
            self: Updated Class
        """
        time, SCurve = self.UnitHydro.index, self.SUnitHydro
        new_time = np.arange(time[0], time[-1]+duration, duration)
        interp_func = interp1d(time, SCurve.values, fill_value='extrapolate',
                               **interp_kwargs)
        SCurve_new = pd.Series(interp_func(new_time), index=new_time)
        uh_new = (SCurve_new-SCurve_new.shift(1).fillna(0))
        uh_new = uh_new.where(uh_new > 0).dropna()
        uh_new.loc[uh_new.index[-1]+duration] = 0
        uh_new.loc[0] = 0
        uh_new = uh_new.sort_index()

        # Ensure that the unit hydrograph acummulates a volume of 1mm
        volume = np.trapz(uh_new, uh_new.index*3600)
        volume = volume/self.geoparams['area']/1e3  # mm
        uh_new = uh_new/volume

        params_new = pd.Series([uh_new.max(), uh_new.idxmax(), duration],
                               index=['qpeak', 'tpeak', 'tstep'])

        # Update
        self.timestep = duration
        self.UnitHydro = uh_new
        self.params = params_new
        self.SUnitHydro = SCurve_new
        return self

    def convolve(self, rainfall: Union[pd.Series, pd.DataFrame], **kwargs: Any
                 ) -> Union[pd.Series, pd.DataFrame]:
        """
        Solve for the convolution of a rainfall time series and the
        unit hydrograph

        Args:
            rainfall (array_like): Series of rain data

        Raises:
            ValueError: If the unit hydrograph is not computed inside the class

        Returns:
            (pandas.Series): flood hydrograph 
        """
        dt = rainfall.index[1]-rainfall.index[0]
        if dt != self.timestep:
            text = 'Rain series and UH time resolution must match !!'
            raise RuntimeError(text)
        if len(rainfall.shape) > 1:
            def func(col): return sg.convolve(col, self.UnitHydro)
            hydrograph = rainfall.apply(func, **kwargs)
        else:
            hydrograph = pd.Series(sg.convolve(rainfall, self.UnitHydro))
        hydrograph.index = hydrograph.index*self.timestep
        return hydrograph

    def compute(self, timestep: float, upper_tail_threshold: float = 1e-4,
                **kwargs: Any) -> Type['LumpedUnitHydrograph']:
        """
        Trigger calculation of desired unit hydrograph

        Args:
            timestep (float): Desired time step in hours.
            upper_tail_threshold (float, optional): Tolerance for the unit
                hydrograph upper tail volume accumulation. High tolerance could
                imply a unit  hydrograph that isnt strighly unitary.
                Defaults to 1e-4. 

        Raises:
            ValueError: If give the class a wrong UH kind.

        Returns:
            self: Updated Class
        """
        self.kwargs = {**self.kwargs, **kwargs}
        self.timestep = timestep
        method = self.method
        if method == 'Clark':
            uh, uh_params = self._Clark(timestep=timestep, **kwargs)
        elif method == 'SCS':
            uh, uh_params = self._SCS(timestep=timestep, **kwargs)
        elif method == 'Linsley':
            uh, uh_params = self._Linsley(**kwargs)
        elif method == 'Gray':
            uh, uh_params = self._Gray(**kwargs)
        else:
            raise ValueError(f'method="{method}" not valid!')

        uh = uh[(uh.cumsum()/uh.sum()) < 1-upper_tail_threshold]
        self.UnitHydro, self.params = uh, uh_params
        self.SUnitHydro = self.get_SHydrograph()
        self.update_duration(self.timestep)

        return self

    def plot(self, **kwargs: Any
             ) -> Tuple[matplotlib.figure.Figure, matplotlib.axes.Axes]:
        """
        Simple accessor to plotting the unit hydrograph
        Args:
            **kwargs are given to pandas plot method
        Returns:
            output of pandas plot method
        """
        params = self.params.to_dict()
        text = [f'{key}: {val:.2f}' for key, val in params.items()]
        text = ' ; '.join(text)
        fig = self.UnitHydro.plot(xlabel='(hr)', ylabel='m3 s-1 mm-1',
                                  title=text, **kwargs)
        ax = matplotlib.pyplot.gca()
        return (fig, ax)



'''
 Author: Lucas Glasner (lgvivanco96@gmail.com)
 Create Time: 1969-12-31 21:00:00
 Modified by: Lucas Glasner,
 Modified time: 2024-05-06 16:24:28
 Description:
 Dependencies:
'''

import os
import numpy as np
import pandas as pd
import warnings
import copy as pycopy
import xarray as xr

from typing import Union, Any, Type
from numpy.typing import ArrayLike
from .abstractions import SCS_Abstractions
from .global_vars import SHYETO_DATA
from .misc import obj_to_xarray

import scipy.stats as st

# ----------------------- duration coefficient routines ---------------------- #


def grunsky_coef(storm_duration: Union[int, float],
                 ref_duration: Union[int, float] = 24) -> float:
    """
    This function computes the duration coefficient
    given by the Grunsky Formula.

    References:
        ???

    Args:
        storm_duration (array_like): storm duration in (hours)
        ref_duration (array_like): Reference duration (hours).
            Defaults to 24 hr

    Returns:
        CD (array_like): Duration coefficient in (dimensionless)
    """
    CD = np.sqrt(storm_duration/ref_duration)
    return CD


def bell_coef(storm_duration: Union[int, float],
              ref_duration: Union[int, float] = 24) -> float:
    """
    This function computes the duration coefficient
    given by the Bell Formula.

    References:
        Bell, F.C. (1969) Generalized pr-Duration-Frequency
        Relationships. Journal of Hydraulic Division, ASCE, 95, 311-327.

    Args:
        storm_duration (array_like): duration in (hours)

    Returns:
        CD (array_like): Duration coefficient in (dimensionless)
    """
    a = (0.54*((storm_duration*60)**0.25)-0.5)
    b = grunsky_coef(1, ref_duration)
    CD = a*b
    return CD


def duration_coef(storm_duration: Union[int, float],
                  ref_duration: Union[int, float] = 24,
                  bell_threshold: Union[int, float] = 1,
                  duration_threshold: Union[int, float] = 10/60) -> float:
    """
    The duration coefficient is a parameter used to transform a known duration
    precipitation to a new duration rain. For example it can be used to
    estimate from daily rainfall (24 hours) the expected accumulation in
    6 hours. This function uses a merge of Grunsky and Bell Formulations of the
    duration coefficient. The idea is to use Bell's Formula only when the input
    duration is less than a user specified threshold. In addition, when the
    duration is less than the "duration_threshold" the duration is set to the
    "duration_threshold".

    References:
        Bell, F.C. (1969) Generalized Rainfall-Duration-Frequency
        Relationships. Journal of Hydraulic Division, ASCE, 95, 311-327.

        Grunsky (???)


    Args:
        storm_duration (array_like): Storm duration in hours
        bell_threshold (float, optional): Duration threshold for changing
            between Grunsky and Bell formulas. Defaults to 1 (hour).
        duration_threshold (float, optional): Minimum storm duration.
            Defaults to 10 minutes (1/6 hours).

    Returns:
        coef (array_like): Duration coefficients (dimensionless)
    """
    if np.isscalar(storm_duration):
        storm_duration = np.array([storm_duration])
    coefs = np.full(storm_duration.shape, np.nan)
    duration_mask = storm_duration < duration_threshold
    bell_mask = storm_duration < bell_threshold
    if duration_mask.sum() != 0:
        threshold = f'{duration_threshold*60:.1f}'
        text = f'A storm duration is less than {threshold} min threshold,'
        text = text+f' setting to {threshold} min.'
        warnings.warn(text)
        storm_duration[duration_mask] = duration_threshold
    coefs[bell_mask] = bell_coef(storm_duration[bell_mask],
                                 ref_duration=ref_duration)
    coefs[~bell_mask] = grunsky_coef(storm_duration[~bell_mask],
                                     ref_duration=ref_duration)
    return coefs


# ------------------------------- Design Storms ------------------------------ #


class RainStorm(object):
    """
    RainStorm class used to building temporal rainfall distributions. 
    The class can be used to build rainstorms that follow any of scipy
    theoretical distributions (e.g 'norm', 'skewnorm', 'gamma', etc) or 
    the empirical rain distributions of the SCS type I, IA, II, III and the 
    Chilean synthetic hyetographs of (Espildora and Echavarría 1979),
    (Benitez and Verni 1985) and (Varas 1985). 

    Examples:
        + Distribute a 24 hour 100 mm rainstorm in a 12 hour gaussian pulse
        -> storm = RainStorm('norm')
        -> storm = storm.compute(timestep=0.5, duration=12, rainfall=100)
        -> storm.pr.plot()

        + Create a 24 hour storm following the SCS type I hyetograph with 
        + pulses every 10 minutes and a total precipitation of 75 mm.
        + Then compute infiltration using SCS method and a basin CN of 75
        -> storm = RainStorm('SCS_I24')
        -> storm = storm.compute(timestep=10/60, duration=24, rainfall=75)
        -> storm = storm.infiltrate(method='SCS', cn=75)
        -> storm.pr.plot()
        -> storm.losses.plot()

        + Create a narrow and wide gaussian pulse of 100 mm in 12 hours
        -> narrow = RainStorm('norm', loc=0.5, scale=0.05)
        -> wide   = RainStorm('norm', loc=0.5, scale=0.15)
        -> narrow = storm.compute(timestep=0.5, duration=12, rainfall=100)
        -> wide   = storm.compute(timestep=0.5, duration=12, rainfall=100)
    """

    def synth_rain(self,
                   loc: float, scale: float, flip: bool = False, n: int = 1000,
                   **kwargs: Any) -> pd.Series:
        """
        Synthetic hyetograph generator function. If the storm type given
        in the class constructor is part of any of scipy distributions 
        the synthetic hyetograph will be built with the given loc, scale
        and scipy default parameters. 

        Args:
            loc (float, optional): Location parameter for distribution type
                hyetographs. Defaults to 0.5.
            scale (float, optional): Scale parameter for distribution type
                hyetographs. Defaults to 0.1.
            flip (bool): Whether to flip the distribution along the x-axis
                or not. Defaults to False.
            n (int, optional): Number of records in the dimensionless storm
            **kwargs are given to scipy.rv_continuous.pdf

        Returns:
            (pandas.Series): Synthetic Hyetograph 1D Table
        """
        time_dimless = np.linspace(0, 1, n)
        kind = self.kind
        scipy_distrs = [d for d in dir(st)
                        if isinstance(getattr(st, d), st.rv_continuous)]
        if kind in scipy_distrs:
            distr = eval(f'st.{kind}')
            shyeto = distr.pdf(time_dimless, loc=loc, scale=scale,
                               **kwargs)
            shyeto = shyeto/np.sum(shyeto)
            if flip:
                shyeto = pd.Series(shyeto[::-1], index=time_dimless)
            else:
                shyeto = pd.Series(shyeto, index=time_dimless)
        else:
            shyeto = SHYETO_DATA[kind]
        return shyeto

    def __init__(self, kind: str = 'norm', loc: float = 0.5, scale: float = 0.1,
                 **kwargs: Any) -> None:
        """
        Synthetic RainStorm builder

        Args:
            kind (str): Type of synthetic hyetograph to use. It can be of two
                types:
                    1) Any of scipy distributiosn (give parameters in **kwargs)
                    2) Any of
                        "SCS_X" with X = I24,IA24,II6,II12,II24,II48,III24
                        "GX_Benitez1985" with X = 1,2,3
                        "GX_Espildora1979" with X = 1,2,3
                        "GXpY_Varas1985" with X = 1,2,3,4 and Y=10,25,50,75,90
                Defaults to 'norm'.
            loc (float): Number between 0 - 1 to specify location parameter
                for statistic-like rainfall distribution. Defaults to 0.5.
            scale (float): Number between 0 -1 to specify scale parameter
                for statistic-like rainfall distribution. Defaults to 0.1.
            **kwargs are given to scipy.rv_continuous.pdf

        Examples:
            RainStorm('SCS_I24')
            RainStorm('G2_Benitez1985')
            RainStorm('G3_Espildora1979')
            RainStorm('G4p10_Varas1985')
            RainStorm('norm', loc=0.5, scale=0.2)
            RainStorm('gamma', loc=0, scale=0.15, a=2)
        """

        self.kind = kind
        self.timestep = None
        self.duration = None
        self.rainfall = None
        self.infiltration = None

        self.pr = None
        self.pr_eff = None
        self.losses = None
        self.pr_dimless = self.synth_rain(loc=loc, scale=scale, **kwargs)

    def __repr__(self) -> str:
        """
        What to show when invoking a RainStorm object
        Returns:
            str: Some metadata
        """
        text = f"RainStorm(kind='{self.kind}', timestep={self.timestep}, "
        text = text+f"duration={self.duration}, "
        text = text+f"infiltration='{self.infiltration}')"
        return text

    def copy(self) -> Type['RainStorm']:
        """
        Create a deep copy of the class itself
        """
        return pycopy.deepcopy(self)

    def compute(self, timestep: Union[int, float], duration: Union[int, float],
                rainfall: ArrayLike, n: int = 1,
                interp_kwargs: dict = {'method': 'linear'}
                ) -> Type['RainStorm']:
        """
        Trigger computation of design storm for a given timestep, storm 
        duration, and total precipitation.

        Args:
            timestep (float): Storm timestep or resolution in hours
            duration (float): Storm duration in hours
            rainfall (array_like or float): Total precipitation in mm. 
            n (int, optional): If n=1 the storm time length will be equal to 
                the user storm duration. If n>1 it will fill the time index with
                n zeros. Defaults to 1.
            interp_kwargs (dict): extra arguments for the interpolation function

        Returns:
            Updated class
        """
        self.timestep = timestep
        self.duration = duration
        self.rainfall = rainfall

        xr_rainfall = obj_to_xarray(rainfall).squeeze()
        dims = {dim: xr_rainfall[dim].shape[0] for dim in xr_rainfall.dims}
        time1 = np.arange(0, duration, timestep)
        time2 = np.arange(0, duration+timestep, timestep)
        time3 = np.arange(0, duration+n*timestep, timestep)

        # Build dimensionless storm (accumulates 1 mm)
        shyeto = obj_to_xarray(self.pr_dimless.cumsum(), dims=('time'),
                               coords={'time': self.pr_dimless.index})
        shyeto = shyeto.interp(coords={'time': np.linspace(0, 1, len(time1))},
                               **interp_kwargs)
        shyeto.coords['time'] = time1

        # Build real storm for the given rainfall
        storm = shyeto.expand_dims(dim=dims)*xr_rainfall
        storm = storm.reindex({'time': time2}).shift({'time': 1})
        storm = storm.diff('time').transpose(*(['time']+list(dims.keys())))
        storm = storm.where(storm >= 0).fillna(0)
        storm = storm.reindex({'time': time3}).fillna(0)
        storm.name = 'pr'

        self.pr = storm
        return self

    def infiltrate(self, method: str = 'SCS', **kwargs: Any
                   ) -> Type['RainStorm']:
        """
        Compute losses due to infiltration with different methods for the
        stored storm Hyetograph
        Args:
            method (str, optional): Infiltration routine. Defaults to 'SCS'.

        Returns:
            Updated class
        """
        self.infiltration = method
        storm = self.pr
        if method == 'SCS':
            cn = kwargs['cn']
            kwargs = kwargs.copy()
            kwargs.pop('cn', None)

            storm_cum = storm.cumsum('time')
            time = np.arange(0, self.duration+self.timestep, self.timestep)
            losses = xr.apply_ufunc(SCS_Abstractions, storm_cum, cn,
                                    kwargs=kwargs,
                                    input_core_dims=[['time'], []],
                                    output_core_dims=[['time']],
                                    vectorize=True)
            losses = losses.reindex({'time': time})
            losses = losses.transpose(*storm.dims).diff('time')
            losses = losses.where(losses >= 0).fillna(0)

            pr_eff = self.pr-losses
            pr_eff = pr_eff.where(pr_eff >= 0).fillna(0)

            self.losses = losses
            self.pr_eff = pr_eff
        else:
            raise ValueError(f'{method} unknown infiltration method.')
        return self



'''
 Author: Lucas Glasner (lgvivanco96@gmail.com)
 Create Time: 1969-12-31 21:00:00
 Modified by: Lucas Glasner, 
 Modified time: 2024-05-06 16:59:31
 Description:
 Dependencies:
'''

import os
import pandas as pd
import numpy as np

import geopandas as gpd
import rioxarray as rxr
import xarray as xr

from osgeo import gdal, gdal_array
from rasterio.features import shapes
from shapely.geometry import shape

from typing import Union, Tuple, Any
from numpy.typing import ArrayLike
from .global_vars import GDAL_EXCEPTIONS

if GDAL_EXCEPTIONS:
    gdal.UseExceptions()
else:
    gdal.DontUseExceptions()

# ------------------------------------ gis ----------------------------------- #


def polygonize(da: xr.DataArray, filter_areas: float = 0) -> gpd.GeoDataFrame:
    """
    Polygonize a boolean rioxarray raster
    Args:
        da (xarray.DataArray): Loaded raster as an xarray object. This should
            have typical rioxarray attributes like da.rio.crs
            and da.rio.transform
        filter_areas (float, optional): Remove polygons with an area less than
            filter areas. Defaults to 0.

    Returns:
        (geopandas.GeoDataFrame): polygonized boolean raster
    """
    da = da.astype(int)
    polygons = []
    for s, v in shapes(da, transform=da.rio.transform()):
        if v == 1:
            polygons.append(shape(s))
    polygons = gpd.GeoSeries(polygons)
    polygons = gpd.GeoDataFrame(geometry=polygons, crs=da.rio.crs)
    polygons = polygons.where(polygons.area > filter_areas).dropna()
    return polygons


def obj_to_xarray(obj: ArrayLike,
                  **kwargs: Any) -> Union[xr.DataArray, xr.Dataset]:
    """
    This function recieves an object and transform it to an xarray object.
    Only accepts pandas series, dataframes, numpy arrays, lists, tuples, or
    array like objects.

    Args:
        obj (ArrayLike): input n-dimensional array.
        **kwargs (optional): Additional keyword arguments to pass to
            xarray.DataArray constructor

    Raises:
        RuntimeError: If input object is not a pandas series, dataframe, 
            list, tuple, int, float, xarray or numpy object.

    Returns:
        (xarray): Transformed array into an xarray style.
    """
    if isinstance(obj, np.ndarray):
        new_xarray = xr.DataArray(obj, **kwargs)
    elif isinstance(obj, pd.Series) or isinstance(obj, pd.DataFrame):
        return obj_to_xarray(obj.values, **kwargs)
    elif isinstance(obj, list) or isinstance(obj, tuple):
        return obj_to_xarray(np.array(obj), **kwargs)
    elif isinstance(obj, int) or isinstance(obj, float):
        return obj_to_xarray(np.array([obj]), **kwargs)
    elif isinstance(obj, xr.DataArray) or isinstance(obj, xr.Dataset):
        new_xarray = obj
    else:
        text = f'Only ints, floats, lists, tuples, pandas, numpy or'
        text = text + f'xarray objects can be used. Got {type(obj)} instead.'
        raise RuntimeError(text)
    return new_xarray


def xarray2gdal(da: xr.DataArray, nodata: Union[str, int, float] = None
                ) -> gdal.Dataset:
    """
    Convert an xarray dataset to a GDAL in-memory dataset.

    Args:
        da (xarray.DataArray): Input data array.
        nodata (_type_, optional): Input dtype. Defaults to None.

    Returns:
        (osgeo.gdal.Dataset): GDAL in-memory dataset
    """
    # Get the data and metadata
    data = da.values
    meta = da.rio.crs.to_wkt()
    transform = da.rio.transform()
    if nodata is None:
        nodata = da.rio.nodata
    dtype = gdal_array.NumericTypeCodeToGDALTypeCode(data.dtype)

    # Write the data to a GDAL in-memory dataset
    driver = gdal.GetDriverByName('MEM')
    outRaster = driver.Create('', da.sizes['x'], da.sizes['y'], 1, dtype)

    # Set geotransform, projection and inner data
    outRaster.SetGeoTransform(transform.to_gdal())
    outRaster.SetProjection(meta)
    outRaster.GetRasterBand(1).WriteArray(data)
    return outRaster


def get_coordinates_from_gdal(gdal_ds: gdal.Dataset
                              ) -> Tuple[np.ndarray, np.ndarray]:
    """
    Get x, y coordinates from a GDAL in-memory dataset.

    Args:
        gdal_ds (osgeo.gdal.Dataset): GDAL in-memory dataset.

    Returns:
        (np.ndarray, np.ndarray): Tuple with arrays of x,y coordinates.
    """
    # Get geotransform
    transform = gdal_ds.GetGeoTransform()

    # Get raster size
    cols = gdal_ds.RasterXSize
    rows = gdal_ds.RasterYSize

    # Calculate x coordinates (for each column)
    x_coords = np.array([transform[0] + j * transform[1] for j in range(cols)])

    # Calculate y coordinates (for each row)
    y_coords = np.array([transform[3] + i * transform[5] for i in range(rows)])

    return x_coords, y_coords


def gdal2xarray(gdal_ds: gdal.Dataset) -> xr.DataArray:
    """
    Convert a GDAL in-memory dataset to an xarray dataset.

    Args:
        gdal_ds (osgeo.gdal.Dataset): Input GDAL in-memory dataset.

    Returns:
        (xarray.DataArray): Output xarray data array.
    """
    # Get data and metadata
    data = gdal_ds.GetRasterBand(1).ReadAsArray()
    transform = rxr.raster_dataset.Affine(*gdal_ds.GetGeoTransform())
    crs = rxr.crs.CRS(gdal_ds.GetProjection())
    nodata = gdal_ds.GetRasterBand(1).GetNoDataValue()

    # Get raster coordinates
    x, y = get_coordinates_from_gdal(gdal_ds)

    # Write the data to an xarray dataarray
    da = xr.DataArray(data, dims=['y', 'x'], coords={'y': y, 'x': x})
    da = da.rio.set_spatial_dims(x_dim='x', y_dim='y')
    da = da.rio.write_crs(crs).rio.write_transform(transform)
    da.rio.update_attrs({'_FillValue': nodata})
    return da


def chile_region(point: gpd.GeoSeries, regions: gpd.GeoDataFrame) -> Any:
    """
    Given a geopandas point and region/states polygones
    this function returns the ID of the polygon where the point belongs
    (as long as an ID column exists in the region polygons)
    If an ID column doesnt exist in the region polygons this function 
    will return the polygon index.

    Args:
        point (geopandas.Series): point 
        regions (geopandas.GeoDataFrame): collection of polygons that
            represents regions or states.

    Returns:
        (Any): ID of the polygon that contains the point
    """
    if 'ID' not in regions.columns:
        regions['ID'] = regions.index

    point = point.to_crs(regions.crs)
    mask = []
    for i in range(len(regions)):
        geom = regions.geometry.iloc[i]
        mask.append(point.within(geom))
    region = regions.iloc[mask, :]['ID']
    return region.item()


def raster_distribution(raster: xr.DataArray, **kwargs: Any) -> pd.Series:
    """
    Given a raster this function computes the histogram

    Args:
        raster (xarray.DataArray): raster

    Returns:
        (pandas.Series): Histogram with data in index and pdf in values
    """
    total_pixels = raster.size
    total_pixels = total_pixels-np.isnan(raster).sum()
    total_pixels = total_pixels.item()

    values = raster.squeeze().values.flatten()
    values = values[~np.isnan(values)]
    dist, values = np.histogram(values, **kwargs)
    dist, values = dist/total_pixels, 0.5*(values[:-1]+values[1:])
    return pd.Series(dist, index=values, name=f'{raster.name}_dist')

# ----------------------------- resources access ----------------------------- #


def load_swampy_data() -> Tuple[gpd.GeoDataFrame, gpd.GeoDataFrame,
                                xr.DataArray, xr.DataArray]:
    """
    Load base example data
    Returns:
        (tuple): (basin polygon, river network segments, dem, curve_number)
    """
    root_folder = os.path.dirname(os.path.abspath(__file__))
    data_folder = os.path.join(root_folder, 'resources', 'RioGomez')
    basin_path = os.path.join(data_folder, 'Cuenca_RioGomez_v0.shp')
    rivers_path = os.path.join(data_folder, 'rnetwork_RioGomez_v0.shp')
    dem_path = os.path.join(data_folder, 'DEM_ALOSPALSAR_RioGomez_v0.tif')
    cn_path = os.path.join(data_folder, 'CurveNumber_RioGomez_v0.tif')

    basin = gpd.read_file(basin_path)
    rivers = gpd.read_file(rivers_path)
    dem = rxr.open_rasterio(dem_path, masked=True).squeeze()
    cn = rxr.open_rasterio(cn_path, masked=True).squeeze()
    dem.name = 'elevation'
    cn.name = 'cn'
    return (basin, rivers, dem, cn)


def load_steep_data() -> Tuple[gpd.GeoDataFrame, gpd.GeoDataFrame,
                               xr.DataArray, xr.DataArray]:
    """
    Load base example data
    Returns:
        (tuple): (basin polygon, river network segments, dem, curve_number)
    """
    root_folder = os.path.dirname(os.path.abspath(__file__))
    data_folder = os.path.join(root_folder, 'resources', 'CNT2420_2')
    basin_path = os.path.join(data_folder, 'CN-T-2420_2.shp')
    rivers_path = os.path.join(data_folder, 'rnetwork.shp')
    dem_path = os.path.join(data_folder, 'dem.tif')
    cn_path = os.path.join(data_folder, 'cn.tif')

    basin = gpd.read_file(basin_path)
    rivers = gpd.read_file(rivers_path)
    dem = rxr.open_rasterio(dem_path, masked=True).squeeze()
    cn = rxr.open_rasterio(cn_path, masked=True).squeeze()
    dem.name = 'elevation'
    cn.name = 'cn'
    return (basin, rivers, dem, cn)


def load_urban_data() -> Tuple[gpd.GeoDataFrame, gpd.GeoDataFrame,
                               xr.DataArray, xr.DataArray]:
    """
    Load base example data
    Returns:
        (tuple): (basin polygon, river network segments, dem, curve_number)
    """
    root_folder = os.path.dirname(os.path.abspath(__file__))
    data_folder = os.path.join(root_folder, 'resources', 'EsteroVDM')
    basin_path = os.path.join(data_folder, 'EsteroVDM.gpkg')
    rivers_path = os.path.join(data_folder, 'rnetwork.gpkg')
    dem_path = os.path.join(data_folder, 'dem.tif')
    cn_path = os.path.join(data_folder, 'cn.tif')

    basin = gpd.read_file(basin_path)
    rivers = gpd.read_file(rivers_path)
    dem = rxr.open_rasterio(dem_path, masked=True).squeeze()
    cn = rxr.open_rasterio(cn_path, masked=True).squeeze()
    dem.name = 'elevation'
    cn.name = 'cn'
    return (basin, rivers, dem, cn)


def load_example_data(kind: str = 'swampy') -> Tuple[gpd.GeoDataFrame,
                                                     gpd.GeoDataFrame,
                                                     xr.DataArray,
                                                     xr.DataArray]:
    """
    Load base example data
    Returns:
        (tuple): (basin polygon, river network segments, dem, curve_number)
    """
    try:
        import geopandas as gpd
        import rioxarray as rxr
    except Exception as e:
        raise RuntimeError(e)

    if kind == 'swampy':
        return load_swampy_data()
    elif kind == 'steep':
        return load_steep_data()
    elif kind == 'urban':
        return load_urban_data()
    else:
        text = f"kind = {kind}, expected one of: 'swampy', 'steep', 'urban'"
        raise RuntimeError(text)

# ----------------------------------- other ---------------------------------- #


def get_psep() -> str:
    """
    Return path separator for the current os

    Returns:
        (str): path separator
    """
    if os.name == 'nt':
        psep = '\\'
    else:
        psep = '/'
    return psep


def is_iterable(obj: Any) -> bool:
    """
    Simple function to check if an object
    is an iterable.

    Args:
        obj (Any): Any python class

    Returns:
        (bool): True if iterable False if not
    """
    try:
        iter(obj)
        return True
    except:
        return False


def to_numeric(obj: Any) -> Any:
    """
    Simple function to transform an object to a numeric type if possible.

    Args:
        obj (Any): The object to be converted to a numeric type.

    Returns:
        (Any): The converted numeric object if conversion is possible,
            otherwise the original object.
    """
    try:
        return pd.to_numeric(obj)
    except:
        return obj



'''
 Author: Lucas Glasner (lgvivanco96@gmail.com)
 Create Time: 2024-11-03 21:34:05
 Modified by: Lucas Glasner, 
 Modified time: 2024-11-06 09:56:20
 Description:
 Dependencies:
'''

import os
import pandas as pd
import geopandas as gpd

try:
    from . import wb_tools
except:
    _has_whitebox = False
else:
    _has_whitebox = True

# ----------------------------------- PATHS ---------------------------------- #
ROOT_FOLDER = os.path.dirname(os.path.abspath(__file__))
DATA_FOLDER = os.path.join(ROOT_FOLDER, 'resources')
SHYETO_PATH = os.path.join(DATA_FOLDER, 'synthetic_storms.csv')
AB_ZONEPATH = os.path.join(DATA_FOLDER, 'vector', 'ZONAS_LINSLEYCHILE.shp')
CHILE_GRAYZONEPATH = os.path.join(DATA_FOLDER, 'vector', 'ZONAS_GRAYCHILE.shp')

# ---------------------------- DATA AND PARAMETERS --------------------------- #
SHYETO_DATA = pd.read_csv(SHYETO_PATH, index_col=0)

CHILE_UH_LINSLEYPOLYGONS = gpd.read_file(AB_ZONEPATH)
CHILE_UH_LINSLEYPARAMS = {
    'C_t': [0.323,   0.584,   1.351,   0.386],
    'n_t': [0.422,	0.327,   0.237,   0.397],
    'C_p': [144.141, 522.514, 172.775, 355.2],
    'n_p': [-0.796, -1.511,  -0.835,  -1.22],
    'C_b': [5.377,	1.822,	 5.428,	  2.7],
    'n_b': [0.805,	1.412,	 0.717,	  1.104]
}
CHILE_UH_LINSLEYPARAMS = pd.DataFrame(CHILE_UH_LINSLEYPARAMS).T
CHILE_UH_LINSLEYPARAMS.columns = ['I', 'II', 'III', 'IV']

CHILE_UH_GRAYPOLYGONS = gpd.read_file(CHILE_GRAYZONEPATH)
CHILE_UH_GRAYPARAMS = {'a': 24.48, 'b': 0.155}
GDAL_EXCEPTIONS = True



'''
 Author: Lucas Glasner (lgvivanco96@gmail.com)
 Create Time: 1969-12-31 21:00:00
 Modified by: Lucas Glasner,
 Modified time: 2024-05-06 16:40:13
 Description:
 Dependencies:
'''

import pandas as pd
import numpy as np
from numpy.typing import ArrayLike
from typing import Union
# ------------------------ SCS curve number equations ------------------------ #


def cn_correction(cn_II: Union[int, float, ArrayLike],
                  amc: str) -> Union[float, ArrayLike]:
    """
    This function changes the curve number value according to antecedent
    moisture conditions (amc)...

    Args:
        cn_II (int|float|ArrayLike): curve number under normal condition
        amc (str): Antecedent moisture condition.
            Options: 'dry'|'I', 'wet'|'III' or 'normal'|'II'

    Raises:
        RuntimeError: If amc is different than 'dry', 'I', 'wet', 'III' or
            'normal', 'II'. 

    Returns:
        (float): adjusted curve number for given AMC

    Reference:
        Ven Te Chow (1988), Applied Hydrology. MCGrow-Hill
        Soil Conservation Service, Urban hydrology for small watersheds,
        tech. re/. No. 55, U. S. Dept. of Agriculture, Washington, D.E:., 1975.
    """
    if (amc == 'dry') or (amc == 'I'):
        cn_I = 4.2*cn_II/(10-0.058*cn_II)
        return cn_I
    elif (amc == 'normal') or (amc == 'II'):
        return cn_II
    elif (amc == 'wet') or (amc == 'III'):
        cn_III = 23*cn_II/(10+0.13*cn_II)
        return cn_III
    else:
        text = f'amc="{amc}"'
        text = text+' Unkown antecedent moisture condition.'
        raise RuntimeError(text)


def SCS_MaximumRetention(cn: Union[int, float, ArrayLike],
                         cfactor: float = 25.4) -> Union[float, ArrayLike]:
    """
    Calculate SCS soil maximum potential retention.

    Args:
        cn (int|float|ArrayLike): Curve number (dimensionless)
        cfactor (float): Unit conversion factor.
            cfactor = 1 --> inches
            cfactor = 25.4 --> milimeters

    Returns:
        (float): maximum soil retention in mm by default
    """
    S = 1000/cn - 10
    return cfactor*S


def SCS_EquivalentCurveNumber(pr: Union[int, float, ArrayLike],
                              pr_eff: Union[int, float, ArrayLike],
                              r: float = 0.2,
                              cfactor: float = 25.4
                              ) -> Union[float, ArrayLike]:
    """
    Given a rainfall ammount and the related effective precipitation (surface
    runoff) observed, this function computes the equivalent curve number of the
    soil. 

    Args:
        pr (1D array_like or float): Precipitation in mm
        pr_eff (1D array_like or float): Effective precipitation in mm
        r (float, optional): Fraction of the maximum potential retention
            used on the initial abstraction calculation. Defaults to 0.2.
        cfactor (float): Unit conversion factor.
            cfactor = 1 --> inches
            cfactor = 25.4 --> milimeters

    Reference:
        Stowhas, Ludwig (2003). Uso del método de la curva número en cuencas
        heterogéneas. XVI Congreso de la Sociedad Chilena de Ingeniería
        Hidráulica. Pontificia Universidad Católica de Chile. Santiago, Chile.

    Returns:
        (1D array_like or float): Equivalent curve number
    """
    a = r**2
    b = -2*r*pr-pr_eff*(1-r)
    c = pr**2-pr_eff*pr
    S_eq = (-b-(b**2-4*a*c)**0.5)/2/a
    CN_eq = 1000*cfactor/(S_eq+10*cfactor)
    return CN_eq


def SCS_EffectiveRainfall(pr: Union[int, float, ArrayLike],
                          cn: Union[int, float, ArrayLike],
                          r: float = 0.2,
                          **kwargs: float) -> Union[float, ArrayLike]:
    """
    SCS formula for effective precipitation/runoff.

    Args:
        pr (float|array): Precipitation depth [mm]
        cn (float|array): Curve Number [-]
        r (float): Initial abstraction ratio, default 0.2

    Returns:
        float|array: Effective rainfall depth [mm]

    Examples:
        >>> SCS_EffectiveRainfall(50, 80)
        22.89
        >>> SCS_EffectiveRainfall([10,20,30], 75)
        array([0., 2.45, 8.67])
    """
    def _scalar(pr: Union[int, float],
                cn: Union[int, float],
                r: float = 0.2,
                **kwargs: float) -> Union[float]:
        """
        Core SCS calculation for scalar inputs

        Raises:
            ValueError: If pr is not positive
            ValueError: If CN is not between 0 and 100
            ValueError: If r is not positive

        Returns:
            float: Effective rainfall depth [mm]
        """
        if np.isnan(pr) or np.isnan(cn):
            return np.nan
        if pr < 0:
            raise ValueError("Precipitation must be positive")
        if not 0 <= cn <= 100:
            raise ValueError("CN must be between 0 and 100")
        if r <= 0:
            raise ValueError("Initial abstraction ratio must be positive")
        S = SCS_MaximumRetention(cn, **kwargs)
        Ia = r*S
        if pr <= Ia:
            pr_eff = 0
        else:
            pr_eff = (pr-Ia)**2/(pr-Ia+S)
        return pr_eff

    if np.isscalar(pr) and np.isscalar(cn):
        return _scalar(pr, cn, r, **kwargs)
    else:
        return np.vectorize(_scalar, otypes=[float])(pr, cn, r, **kwargs)


def SCS_Abstractions(pr: Union[int, float, ArrayLike],
                     cn: Union[int, float, ArrayLike],
                     r: float = 0.2,
                     **kwargs: float) -> Union[float, ArrayLike]:
    """
    SCS formula for overall water losses due to infiltration/abstraction.
    Losses are computed simply as total precipitation - total runoff. 

    Args:
        pr (array_like or float): Precipitation in mm 
        cn (array_like or float): Curve Number
        r (float, optional): Fraction of the maximum potential retention
            Defaults to 0.2.
        **kwargs are passed to SCS_MaximumRetention function


    Returns:
        (array_like): Losses/Abstraction/Infiltration
    """
    pr_eff = SCS_EffectiveRainfall(pr, cn, r=r, **kwargs)
    Losses = pr-pr_eff
    return Losses



'''
 Author: Lucas Glasner (lgvivanco96@gmail.com)
 Create Time: 1969-12-31 21:00:00
 Modified by: Lucas Glasner, 
 Modified time: 2024-05-06 09:56:20
 Description:
 Dependencies:
'''

import pandas as pd
import numpy as np
import geopandas as gpd
import xarray as xr
import warnings

from typing import Union, Any, Tuple
from numpy.typing import ArrayLike
from shapely.geometry import Point
import networkx as nx

from .abstractions import SCS_MaximumRetention

# ------------------------ Geomorphological properties ----------------------- #


def get_main_river(river_network: Union[gpd.GeoSeries, gpd.GeoDataFrame]
                   ) -> Union[gpd.GeoSeries, gpd.GeoDataFrame]:
    """
    For a given river network (shapefile with river segments) this functions
    creates a graph with the river network and computes the main river with the
    longest_path algorithm. 

    Args:
        river_network (GeoDataFrame): River network (lines)
    Returns:
        (GeoDataFrame): Main river extracted from the river network
    """
    # Get network connectivity information
    river_network = river_network.explode(index_parts=False)
    start_node = river_network.geometry.apply(lambda g: Point(g.coords[0]))
    end_node = river_network.geometry.apply(lambda g: Point(g.coords[-1]))
    weight = river_network.length
    ids = river_network.index

    # Create River Network Graph
    G = nx.DiGraph()
    for n, a, b, w in zip(ids, start_node, end_node, weight):
        G.add_edge(a, b, index=n, weight=w)

    # Get the main river segments
    main_river = nx.dag_longest_path(G)
    mask = start_node.isin(main_river)
    main_river = river_network.loc[mask]
    return main_river


def basin_outlet(basin: Union[gpd.GeoSeries, gpd.GeoDataFrame],
                 dem: xr.DataArray, n: int = 3
                 ) -> Tuple[np.ndarray, np.ndarray]:
    """
    This function computes the basin outlet point defined as the
    point of minimum elevation along the basin boundary.

    Args:
        basin (geopandas.GeoDataFrame): basin polygon
        dem (xarray.DataArray): Digital elevation model
        n (int, optional): Number of DEM pixels to consider for the
            elevation boundary. Defaults to 3.

    Returns:
        outlet_y, outlet_x (tuple): Tuple with defined outlet y and x
            coordinates.
    """
    dx = (max(dem.y.diff('y')[0], dem.x.diff('x')[0])).item()
    basin_boundary = basin.boundary
    dem_boundary = dem.rio.clip(basin_boundary.buffer(dx*n))
    dem_boundary = dem_boundary.where(dem_boundary != -9999)
    outlet_point = dem_boundary.isel(**dem_boundary.argmin(['y', 'x']))
    outlet_y, outlet_x = outlet_point.y.item(), outlet_point.x.item()
    return (outlet_y, outlet_x)


def basin_geographical_params(fid: Union[str, int, float],
                              basin: Union[gpd.GeoSeries, gpd.GeoDataFrame],
                              outlet: ArrayLike = None) -> pd.DataFrame:
    """
    Given a basin id and a basin polygon as a geopandas object 
    this function computes the "geographical" or vector properties of
    the basin (i.e centroid coordinates, area, perimeter and outlet to
    centroid length.)

    Args:
        fid (str|int|float): basin identifier
        basin (geopandas.GeoSeries|geopandas.GeoDataFrame): basin polygon
        outlet (ArrayLike, optional): shape (2,) array with basin x,y outlet
            point. Defaults to None
    Raises:
        RuntimeError: If outlet == None and the basin doesnt have the drainage
            point in the attribute table. (outlet_x and outlet_y columns)

    Returns:
        pandas.DataFrame: table with parameters
    """
    if type(outlet) != type(None):
        outlet_x, outlet_y = (outlet[0], outlet[1])
        basin['outlet_x'] = outlet_x
        basin['outlet_y'] = outlet_y

    if not (('outlet_x' in basin.columns) or ('outlet_y' in basin.columns)):
        error = 'Basin polygon attribute table must have an "outlet_x" and'
        error = error+' "outlet_y" columns.'
        error = error+' If not, use the outlet argument.'
        raise RuntimeError(error)

    params = pd.DataFrame([], index=[fid])
    params['outlet_x'] = basin.outlet_x.item()
    params['outlet_y'] = basin.outlet_y.item()
    params['centroid_x'] = basin.centroid.x.item()
    params['centroid_y'] = basin.centroid.y.item()
    params['area'] = basin.area.item()/1e6
    params['perim'] = basin.boundary.length.item()/1e3

    # Outlet to centroid
    outlet = Point(basin.outlet_x.item(), basin.outlet_y.item())
    out2cen = basin.centroid.distance(outlet)
    params['out2centroidlen'] = out2cen.item()/1e3

    return params


def terrain_exposure(aspect: xr.DataArray, fid: Union[str, int, float] = 0
                     ) -> pd.DataFrame:
    """
    From an aspect raster compute the percentage of the raster that
    belong to each of the 8 typical geographical directions.
    (i.e N, S, E, W, NE, SE, SW, NW).

    Args:
        aspect (xarray.DataArray): Aspect raster
        fid (_type_, optional): Feature ID. Defaults to 0.'

    Returns:
        pandas.DataFrame: Table with main directions exposure
    """
    # Direction of exposure
    direction_ranges = {
        'N_exposure': (337.5, 22.5),
        'S_exposure': (157.5, 202.5),
        'E_exposure': (67.5, 112.5),
        'W_exposure': (247.5, 292.5),
        'NE_exposure': (22.5, 67.5),
        'SE_exposure': (112.5, 157.5),
        'SW_exposure': (202.5, 247.5),
        'NW_exposure': (292.5, 337.5),
    }
    # Calculate percentages for each direction
    tot_pixels = np.size(aspect.values) - \
        np.isnan(aspect.values).sum()
    dir_perc = {}

    for direction, (min_angle, max_angle) in direction_ranges.items():
        if min_angle > max_angle:
            exposure = np.logical_or(
                (aspect.values >= min_angle) & (
                    aspect.values <= 360),
                (aspect.values >= 0) & (aspect.values <= max_angle)
            )
        else:
            exposure = (aspect.values >= min_angle) & (
                aspect.values <= max_angle)

        direction_pixels = np.sum(exposure)
        dir_perc[direction] = (direction_pixels/tot_pixels)
    dir_perc = pd.DataFrame(dir_perc.values(),
                            index=dir_perc.keys(),
                            columns=[fid]).T
    return dir_perc


def basin_terrain_params(fid: Union[str, int, float], dem: xr.DataArray
                         ) -> pd.DataFrame:
    """
    From an identifier (fid) and a digital elevation model (DEM) loaded
    as an xarray object, this function computes the following properties:
    1) Minimum, mean, median and maximum height
    2) Difference between maximum and minimum height
    3) Difference between mean and minimum height
    4) Mean slope if slope is in the dataset
    5) % of the terrain in each of the 8 directions (N,S,W,E,SW,SE,NW,NE)

    Args:
        fid (_type_): basin identifier
        dem (xarray.Dataset): Digital elevation model

    Returns:
        pandas.DataFrame: Table with terrain-derived parameters
    """
    if 'elevation' not in dem.variables:
        text = 'Input dem must be an xarray dataset with an "elevation" \
                variable'
        raise RuntimeError(text)
    params = pd.DataFrame([], index=[fid])

    # Height parameters
    params['hmin'] = dem.elevation.min().item()
    params['hmax'] = dem.elevation.max().item()
    params['hmean'] = dem.elevation.mean().item()
    params['hmed'] = dem.elevation.median().item()
    params['deltaH'] = params['hmax']-params['hmin']
    params['deltaHm'] = params['hmean']-params['hmin']

    # Slope parameters
    if 'slope' in dem.variables:
        params['meanslope'] = dem.slope.mean().item()
    else:
        warnings.warn('"slope" variable doesnt exists in the dataset!')

    # Exposure/Aspect parameters
    if 'aspect' in dem.variables:
        dir_perc = terrain_exposure(dem.aspect, fid=fid)
        params = pd.concat([params, dir_perc], axis=1)
    else:
        warnings.warn('"aspect" variable doesnt exists in the dataset!')
    return params

# -------------------- Concentration time for rural basins ------------------- #


def tc_SCS(mriverlen: Union[int, float], meanslope: Union[int, float],
           curvenumber: Union[int, float], **kwargs: Any) -> float:
    """
    USA Soil Conservation Service (SCS) method.
    Valid for rural basins ¿?.

    Reference:
        Part 630 National Engineering Handbook. Chapter 15. NRCS 
        United States Department of Agriculture.

    Args:
        mriverlen (float): Main river length in (km)
        meanslope (float): Basin mean slope in m/m
        curvenumber (float): Basin curve number (dimensionless)
        **kwargs do nothing

    Returns:
        Tc (float): Concentration time (minutes)
    """
    mriverlen_ft = 3280.84*mriverlen
    potentialstorage_inch = SCS_MaximumRetention(curvenumber, cfactor=1)
    slope_perc = meanslope*100
    numerator = mriverlen_ft**0.8*((potentialstorage_inch+1) ** 0.7)
    denominator = 1140*slope_perc**0.5
    Tc = numerator/denominator*60  # 60 minutes = 1 hour
    return Tc


def tc_kirpich(mriverlen: Union[int, float], hmax: Union[int, float],
               hmin: Union[int, float], **kwargs: Any) -> float:
    """
    Kirpich equation method.
    Valid for small and rural basins ¿?.

    Reference:
        ???

    Args:
        mriverlen (float): Main river length in (km)
        hmax (float): Basin maximum height (m)
        hmin (float): Basin minimum height (m)
        **kwargs do nothing

    Returns:
        Tc (float): Concentration time (minutes)
    """
    deltaheights = hmax-hmin
    Tc = ((1000*mriverlen)**1.15)/(deltaheights**0.385)/51
    return Tc


def tc_giandotti(mriverlen: Union[int, float], hmean: Union[int, float],
                 hmin: Union[int, float], area: Union[int, float],
                 **kwargs: Any) -> float:
    """
    Giandotti equation method.
    Valid for small basins (< 20km2) with high slope (>10%) ¿?. 

    Reference:
        ???

    Args:
        mriverlen (float): Main river length in (km)
        hmean (float): Basin mean height (meters)
        hmin (float): Basin minimum height (meters)
        area (float): Basin area (km2)
        **kwargs do nothing

    Returns:
        Tc (float): Concentration time (minutes)
    """
    a = (4*area**0.5+1.5*mriverlen)
    b = (0.8*(hmean-hmin)**0.5)
    Tc = 60*a/b
    return Tc


def tc_california(mriverlen: Union[int, float], hmax: Union[int, float],
                  hmin: Union[int, float], **kwargs: Any) -> float:
    """
    California Culverts Practice (1942) equation.
    Valid for mountain basins ¿?.

    Reference: 
        ???

    Args:
        mriverlen (float): Main river length in (km)
        hmax (float): Basin maximum height (m)
        hmin (float): Basin minimum height (m)
        **kwargs do nothing

    Returns:
        Tc (float): Concentration time (minutes)

    """
    deltaheights = hmax-hmin
    Tc = 57*(mriverlen**3/deltaheights)**0.385
    return Tc


def tc_spain(mriverlen: Union[int, float], meanslope: Union[int, float],
             **kwargs: Any) -> float:
    """
    Equation of Spanish/Spain regulation.

    Reference:
        ???

    Args:
        mriverlen (float): Main river length in (km)
        meanslope (float): Basin mean slope in m/m
        **kwargs do nothing

    Returns:
        Tc (float): Concentration time (minutes)
    """
    Tc = 18*(mriverlen**0.76)/((meanslope*100)**0.19)
    return Tc


def concentration_time(method: str, **kwargs: Any) -> float:
    """
    General function for computing the concentration time with different
    formulas.

    Args:
        method (str): Concentration time formula:
            Options:
                California, Giandotti, Kirpich, SCS, Spain, 
        **kwargs are given to the respective concentration time formula

    Raises:
        ValueError: If user asks for an unkown method

    Returns:
        (float): Concentration time (minutes)
    """
    if method == 'California':
        return tc_california(**kwargs)
    elif method == 'Giandotti':
        return tc_giandotti(**kwargs)
    elif method == 'Kirpich':
        return tc_kirpich(**kwargs)
    elif method == 'SCS':
        return tc_SCS(**kwargs)
    elif method == 'Spain':
        return tc_spain(**kwargs)
    else:
        raise ValueError(f'"{method}": Unknown tc method!')



'''
 Author: Lucas Glasner (lgvivanco96@gmail.com)
 Create Time: 2024-08-05 11:11:38
 Modified by: Lucas Glasner,
 Modified time: 2024-08-05 11:11:43
 Description: Main watershed classes
 Dependencies:
'''

import os
import warnings


import numpy as np
import pandas as pd
import geopandas as gpd
import rioxarray as rxr
import xarray as xr
import copy as pycopy
import matplotlib
import matplotlib.pyplot as plt

from typing import Union, Any, Optional, Type, Tuple
from osgeo import gdal, gdal_array
from scipy.interpolate import interp1d

from .misc import raster_distribution, polygonize, gdal2xarray, xarray2gdal
from .unithydrographs import LumpedUnitHydrograph as SUH
from .geomorphology import get_main_river, basin_outlet
from .geomorphology import basin_geographical_params, basin_terrain_params
from .global_vars import CHILE_UH_LINSLEYPOLYGONS, CHILE_UH_GRAYPOLYGONS
from .global_vars import GDAL_EXCEPTIONS, _has_whitebox
from .abstractions import cn_correction
from .abstractions import SCS_EffectiveRainfall, SCS_EquivalentCurveNumber


if GDAL_EXCEPTIONS:
    gdal.UseExceptions()
else:
    gdal.DontUseExceptions()
# ---------------------------------------------------------------------------- #


class RiverBasin(object):
    """
    Watershed class used to compute geomorphological properties of basins, 
    unit hydrographs, flood hydrographs, terrain properties, among other 
    hydrological methods. 

    The class seeks to be a virtual representation of an hydrographic basin, 
    where given inputs of terrain, river network and land cover are used to 
    derive basin-wide properties and hydrological computations. 

    Examples:
        + Compute geomorphometric parameters

        -> import geopandas as gpd
        -> import rioxarray as rxr
        -> dem = rxr.open_rasterio('path/to/dem', masked=True)
        -> cn  = rxr.open_rasterio('path/to/cn', masked=True)
        -> basin = gpd.read_file('/path/to/basinpolygon')
        -> rivers = gpd.read_file('/path/to/riversegments/')

        -> wshed = RiverBasin('mybasin', basin, rivers, dem, cn)
        -> wshed.compute_params()

        + Use curve number corrected by a wet/dry condition
        -> wshed = RiverBasin('mybasin', basin, rivers, dem, cn, amc='wet')
        -> wshed.compute_params()

        + Change a parameter by hand
        -> wshed.set_parameter('area', 1000)

        + Check hypsometric curve
        -> curve = wshed.get_hypsometric_curve(bins='auto')

        + Check fraction of area below 1400 meters
        -> fArea = wshed.area_below_height(1400)

        + Get relationship of curve number vs precipitation due to basin land
        + cover heterogeneities. 
        -> cn_curve = wshed.get_equivalent_curvenumber()

        + Access basin params as pandas Data Frame
        -> wshed.params

        + Compute SCS unit hydrograph for rain pulses of 1 hour
        -> wshed.SynthUnitHydro(method='SCS', timestep=1)

        + Compute flood hydrograph with a series of rainfall
        -> whsed.UnitHydro.convolve(rainfall)
    """

    def __init__(self, fid: Union[str, int, float],
                 basin: Union[gpd.GeoSeries, gpd.GeoDataFrame],
                 dem: xr.DataArray,
                 rivers: Union[gpd.GeoSeries, gpd.GeoDataFrame] = None,
                 cn: xr.DataArray = None,
                 amc: str = 'II') -> None:
        """
        This class represents a river basin with associated geographical and
        hydrological data.  It initializes the basin with various attributes
        such as basin identifier, watershed polygon, digital elevation model
        (DEM), river network segments, curve number raster, and antecedent
        moisture condition (AMC).

        Args:
            basin (Union[gpd.GeoSeries, gpd.GeoDataFrame]): Watershed polygon
            dem (xr.DataArray): Digital elevation model
            rivers (Union[gpd.GeoSeries, gpd.GeoDataFrame], optional):
                River network segments. Defaults to None.
            cn (xr.DataArray, optional): Curve Number raster. Defaults to None,
                which leads to an empty curve number raster.
            amc (str, optional): Antecedent moisture condition.
                Defaults to 'II'. Options: - 'dry' or 'I',
                                           - 'normal' or 'II',
                                           - 'wet' or 'III'.
        """
        # ID
        self.fid = fid

        # Vectors
        self.basin = basin.copy()
        self.mask_vector = basin.copy()
        if rivers is not None:
            self.rivers = rivers.copy()
            self.rivers_main = gpd.GeoDataFrame()
        else:
            self.rivers = rivers
            self.rivers_main = gpd.GeoDataFrame()

        # Rasters
        self.dem = dem.rio.write_nodata(-9999).squeeze().copy()
        self.dem = self.dem.to_dataset(name='elevation')
        self.dem.encoding = dem.encoding
        self.mask_raster = ~np.isnan(dem)
        self.mask_raster.name = 'mask'

        if cn is not None:
            self.cn = cn.rio.write_nodata(-9999).squeeze().copy()
            self.cn = cn_correction(self.cn, amc=amc)
            self.cn_counts = pd.DataFrame([])
            self.amc = amc
        else:
            self.cn = cn

        # Properties
        self.params = pd.DataFrame([], index=[self.fid], dtype=object)
        self.hypsometric_curve = pd.Series(dtype=float)
        self.exposure_distribution = pd.Series(dtype=float)
        self.UnitHydro = None

    def __repr__(self) -> str:
        """
        What to show when invoking a RiverBasin object
        Returns:
            str: Some metadata
        """
        if type(self.UnitHydro) != type(None):
            uh_text = self.UnitHydro.method
        else:
            uh_text = None

        if self.params.shape != (1, 0):
            param_text = str(self.params).replace(self.fid, '')
        else:
            param_text = None
        text = f'RiverBasin: {self.fid}\nUnitHydro: {uh_text}\n'
        text = text+f'Parameters: {param_text}'
        return text

    def _process_gdaldem(self, varname: str, **kwargs: Any) -> xr.Dataset:
        """
        Processes a Digital Elevation Model (DEM) using the GDAL DEMProcessing
        utility. This method utilizes the GDAL DEMProcessing command line
        utility to derive various properties from a DEM. The output is returned
        as an xarray Dataset.

        Args:
            varname (str): The name of the DEM derived property to compute.
            **kwargs (Any): Additional keyword arguments to pass to the GDAL
                DEMProcessing function.

        Returns:
            xr.Dataset: An xarray Dataset containing the DEM derived property.
        """
        dem_xr = self.dem.elevation
        dem_gdal = xarray2gdal(dem_xr)

        # Create in-memory output GDAL dataset
        dtype = gdal_array.NumericTypeCodeToGDALTypeCode(dem_xr.dtype)
        mem_driver = gdal.GetDriverByName('MEM')
        out_ds = mem_driver.Create('', dem_xr.sizes['x'], dem_xr.sizes['y'], 1,
                                   dtype)
        out_ds.SetGeoTransform(dem_xr.rio.transform().to_gdal())
        out_ds.SetProjection(dem_xr.rio.crs.to_wkt())

        # Process DEM using gdal.DEMProcessing
        out_ds = gdal.DEMProcessing(out_ds.GetDescription(), dem_gdal, varname,
                                    format='MEM', computeEdges=True, **kwargs)
        out_ds = gdal2xarray(out_ds).to_dataset(name=varname)
        out_ds.coords['y'] = dem_xr.coords['y']
        out_ds.coords['x'] = dem_xr.coords['x']
        return out_ds

    def _processgeography(self, n: int = 3,
                          **kwargs: Any) -> Type['RiverBasin']:
        """
        Compute geographical parameters of the basin

        Args:
            n (int, optional): Number of DEM pixels to consider for the
                elevation boundary. Defaults to 3.
            **kwargs are given to basin_geographical_params function.

        Returns:
            self: updated class
        """
        try:
            c1 = 'outlet_x' not in self.basin.columns
            c2 = 'outlet_y' not in self.basin.columns
            if c1 or c2:
                outlet_y, outlet_x = self._get_basinoutlet(n=n)
            else:
                c3 = self.basin['outlet_x'].item() is None
                c4 = self.basin['outlet_y'].item() is None
                if c3 or c4:
                    outlet_y, outlet_x = self._get_basinoutlet(n=n)

            geo_params = basin_geographical_params(self.fid, self.basin,
                                                   **kwargs)
        except Exception as e:
            geo_params = pd.DataFrame([], index=[self.fid])
            warnings.warn('Geographical Parameters Error:'+f'{e}')
        self.params = pd.concat([self.params, geo_params], axis=1)
        return self

    def _processdem(self, preprocess: bool = True) -> Type['RiverBasin']:
        """
        Processes the Digital Elevation Model (DEM) to compute the hypsometric
        curve, slope, and aspect. Then computes DEM-derived properties for the
        basin and saves them in the params dataframe.
        Args:
            preprocess (bool): If True, preprocess the DEM to compute
                hypsometric curve, slope, and aspect.
            RiverBasin: The updated class instance with computed
                DEM properties.
        """
        try:
            if preprocess:
                curve = self.get_hypsometric_curve()
                slope = self._process_gdaldem('slope', slopeFormat='percent')
                aspect = self._process_gdaldem('aspect')

                slope = slope.where(slope != -9999)
                aspect = aspect.where(aspect != -9999)

                self.dem = xr.merge([self.dem.elevation, slope/100, aspect])
                self.dem.attrs = {
                    'standard_name': 'terrain model',
                    'hypsometry_x': [f'{i:.2f}' for i in curve.index],
                    'hypsometry_y': [f'{j:3f}' for j in curve.values]
                }
            # DEM derived params
            terrain_params = basin_terrain_params(self.fid, self.dem)
            exp = terrain_params.T.index.map(lambda x: 'exposure' in x)
            self.exposure_distribution = terrain_params.T[exp]
        except Exception as e:
            terrain_params = pd.DataFrame([], index=[self.fid])
            warnings.warn('PostProcess DEM Error:'+f'{e}')
        self.params = pd.concat([self.params, terrain_params], axis=1)
        return self

    def _processrivers(self) -> Type['RiverBasin']:
        """
        Compute river network properties

        Returns:
            self: updated class
        """
        try:
            mainriver = get_main_river(self.rivers)
            self.rivers_main = mainriver
            mriverlen = self.rivers_main.length.sum()/1e3
            if mriverlen.item() != 0:
                mriverlen = mriverlen.item()
            else:
                mriverlen = np.nan
            self.params['mriverlen'] = mriverlen
        except Exception as e:
            warnings.warn('Flow derived properties Error: ' + f'{e}')
        return self

    def _processrastercounts(self, raster: xr.DataArray, output_type: int = 1
                             ) -> pd.DataFrame:
        """
        Computes area distributions of rasters (% of the basin area with the
        X raster property)
        Args:
            raster (xarray.DataArray): Raster with basin properties
                (e.g land cover classes, soil types, etc)
            output_type (int, optional): Output type:
                Option 1: 
                    Returns a table with this format:
                    +-------+----------+----------+
                    | INDEX | PROPERTY | FRACTION |
                    +-------+----------+----------+
                    |     0 | A        |          |
                    |     1 | B        |          |
                    |     2 | C        |          |
                    +-------+----------+----------+

                Option 2:
                    Returns a table with this format:
                    +-------------+----------+
                    |    INDEX    | FRACTION |
                    +-------------+----------+
                    | fPROPERTY_A |          |
                    | fPROPERTY_B |          |
                    | fPROPERTY_C |          |
                    +-------------+----------+

                Defaults to 1.
        Returns:
            counts (pandas.DataFrame): Results table
        """
        try:
            counts = raster.to_series().value_counts()
            counts = counts/counts.sum()
            counts.name = self.fid
            if output_type == 1:
                counts = counts.reset_index().rename({self.fid: 'weights'},
                                                     axis=1)
            elif output_type == 2:
                counts.index = [f'f{raster.name}_{i}' for i in counts.index]
                counts = pd.DataFrame(counts)
            else:
                raise RuntimeError(f'{output_type} must only be 1 or 2.')
        except Exception as e:
            counts = pd.DataFrame([], columns=[self.fid],
                                  index=[0])
            warnings.warn('Raster counting Error:'+f'{e}')
        return counts

    def _get_basinoutlet(self, n: int = 3) -> Tuple[np.ndarray, np.ndarray]:
        """
        This function computes the basin outlet point defined as the
        point of minimum elevation along the basin boundary.

        Args:
            n (int, optional): Number of DEM pixels to consider for the
                elevation boundary. Defaults to 3.

        Returns:
            outlet_y, outlet_x (tuple): Tuple with defined outlet y and x
                coordinates.
        """
        outlet_y, outlet_x = basin_outlet(self.basin, self.dem.elevation, n=n)
        self.basin['outlet_x'] = outlet_x
        self.basin['outlet_y'] = outlet_y
        return (outlet_y, outlet_x)

    def copy(self) -> Type['RiverBasin']:
        """
        Create a deep copy of the class itself
        """
        return pycopy.deepcopy(self)

    def set_parameter(self, index: str, value: Any) -> Type['RiverBasin']:
        """
        Simple function to add or fix a parameter to the basin parameters table

        Args:
            index (str): parameter name/id or what to put in the table index
            value (Any): value of the new parameter
        """
        self.params.loc[index, :] = value
        return self

    def get_hypsometric_curve(self, bins: Union[str, int, float] = 'auto',
                              **kwargs: Any) -> pd.Series:
        """
        Compute the hypsometric curve of the basin based on terrain elevation
        data. The hypsometric curve represents the distribution of elevation
        within the basin, expressed as the fraction of the total area that lies
        below a given elevation.
        Args:
            bins (str|int|float, optional): The method or number of
                bins to use for the elevation distribution. Default is 'auto'.
            **kwargs (Any): Additional keyword arguments to pass to the
                raster_distribution function.

        Returns:
            pandas.Series: A pandas Series representing the hypsometric curve,
                where the index corresponds to elevation bins and the values
                represent the cumulative fraction of the area below each
                elevation.
        """
        curve = raster_distribution(self.dem.elevation, bins=bins, **kwargs)
        self.hypsometric_curve = curve.cumsum().drop_duplicates()
        return curve

    def area_below_height(self, height: Union[int, float], **kwargs: Any
                          ) -> float:
        """
        With the hypsometric curve compute the fraction of area below
        a certain height.

        Args:
            height (int|float): elevation value
            **kwargs (Any): Additional keyword arguments to pass to the
                raster_distribution function.

        Returns:
            (float): fraction of area below given elevation
        """
        if len(self.hypsometric_curve) == 0:
            warnings.warn('Computing hypsometric curve ...')
            self.get_hypsometric_curve(**kwargs)
        curve = self.hypsometric_curve
        if height < curve.index.min():
            return 0
        if height > curve.index.max():
            return 1
        interp_func = interp1d(curve.index.values, curve.values)
        return interp_func(height).item()

    def get_equivalent_curvenumber(self,
                                   pr_range: Tuple[float, float] = (1., 1000.),
                                   **kwargs: Any) -> pd.Series:
        """
        Calculate the dependence of the watershed curve number on precipitation
        due to land cover heterogeneities.

        This routine computes the equivalent curve number for a heterogeneous
        basin as a function of precipitation. It takes into account the
        distribution of curve numbers within the basin and the corresponding
        effective rainfall for a range of precipitation values.

        Args:
            pr_range (tuple): Minimum and maximum possible precipitation (mm).
            **kwargs: Additional keyword arguments to pass to the
                SCS_EffectiveRainfall and SCS_EquivalentCurveNumber routine.

        Returns:
            pd.Series: A pandas Series representing the equivalent curve number
                as a function of precipitation, where the index corresponds to
                precipitation values and the values represent the equivalent
                curve number.
        """
        # Precipitation range
        pr = np.linspace(pr_range[0], pr_range[1], 1000)
        pr = np.expand_dims(pr, axis=-1)

        # Curve number counts
        cn_counts = self._processrastercounts(self.cn)
        weights, cn_values = cn_counts['weights'].values, cn_counts['cn'].values
        cn_values = np.expand_dims(cn_values, axis=-1)

        # Broadcast curve number and pr arrays
        broad = np.broadcast_arrays(cn_values, pr.T)

        # Get effective precipitation
        pr_eff = SCS_EffectiveRainfall(pr=broad[1], cn=broad[0], **kwargs)
        pr_eff = (pr_eff.T * weights).sum(axis=-1)

        # Compute equivalent curve number for hetergeneous basin
        curve = SCS_EquivalentCurveNumber(pr[:, 0], pr_eff, **kwargs)
        curve = pd.Series(curve, index=pr[:, 0])
        curve = curve.sort_index()
        self.cn_equivalent = curve
        return curve

    def compute_params(self, preprocess_rivers: bool = False,
                       preprocess_rivers_kwargs: dict = {},
                       dem_kwargs: dict = {},
                       geography_kwargs: dict = {},
                       river_network_kwargs: dict = {}) -> Type['RiverBasin']:
        """

        Compute basin geomorphological properties:
            1) Geographical properties: centroid coordinates, area, etc
                Details in src.geomorphology.basin_geographical_params routine
            2) Terrain properties: DEM derived properties like minimum, maximum
                or mean height, etc.
                Details in src.geomorphology.basin_terrain_params
            3) Flow derived properties: Main river length using graph theory, 
                drainage density and shape factor. 
                Details in src.geomorphology.main_river
        Args:
            preprocess_river_network (bool, optional): Whether to compute 
                river network from given DEM. Requeries whitebox_workflows
                package. Defaults to False. 
            dem_kwargs (dict, optional): Additional arguments for the terrain
                preprocessing function. Defaults to {}.
            geography_kwargs (dict, optional): Additional arguments for the
                geography preprocessing routine. Defauts to {}.
            river_network_kwargs (dict, optional): Additional arguments for the
                main river finding routine. Defaults to {}.
                Details in src.geomorphology.main_river routine
        Returns:
            self: updated class
        """
        if self.params.shape != (1, 0):
            self.params = pd.DataFrame([], index=[self.fid], dtype=object)

        # Geographical parameters
        self._processgeography(**geography_kwargs)

        # Compute slope and aspect. Update dem property
        self._processdem(**dem_kwargs)

        # Flow derived params
        if self.rivers is not None:
            self._processrivers(**river_network_kwargs)
        else:
            if preprocess_rivers:
                from .wb_tools import wbDEMpreprocess
                flow, rivers = wbDEMpreprocess(self.dem.elevation,
                                               return_streams=True,
                                               **preprocess_rivers_kwargs)
                self.dem = xr.merge([self.dem, flow])
                self.rivers = rivers
                self._processrivers(**river_network_kwargs)
                self.params['mriverlen'] = flow.flowlen.max().item()/1e3

        # Curve number process
        if self.cn is not None:
            self.params['curvenumber'] = self.cn.mean().item()

        # Reorder
        self.params = self.params.T.astype(object)

        return self

    def clip(self, polygon: Union[gpd.GeoSeries, gpd.GeoDataFrame],
             **kwargs: Any) -> Type['RiverBasin']:
        """
        Clip watershed data to a specified polygon boundary and create a new
        RiverBasin object. This method creates a new RiverBasin instance with
        all data (basin boundary, rivers, DEM, etc) clipped to the given
        polygon boundary. It also recomputes all geomorphometric parameters for
        the clipped area.

        Args:
            polygon (Union[gpd.GeoSeries, gpd.GeoDataFrame]): Polygon defining
                the clip boundary. Must be in the same coordinate reference
                system (CRS) as the watershed data.
            **kwargs (Any): Additional keyword arguments to pass to
                self.compute_params() method.
        Returns:
            self: A new RiverBasin object containing the clipped data and
                updated parameters.
        Notes:
            - The input polygon will be dissolved to ensure a single boundary
            - No-data values (-9999) are filtered out from DEM and CN rasters
            - All geomorphometric parameters are recomputed for the clipped
              area
        """
        nwshed = self.copy()
        polygon = polygon.dissolve()

        # Basin
        nbasin = self.basin.copy().clip(polygon)
        nwshed.basin = nbasin
        nwshed.mask_vector = nbasin

        # DEM & mask
        ndem = self.dem.copy().rio.clip(polygon.geometry)
        ndem = ndem.where(ndem != -9999)
        ndem = ndem.reindex({'y': self.dem.y, 'x': self.dem.x})
        nmask = ~np.isnan(ndem.elevation)
        nmask.name = 'mask'
        nwshed.dem = ndem
        nwshed.mask_raster = nmask

        # Rivers
        if self.rivers is not None:
            nrivers = self.rivers.copy().clip(polygon)
            nwshed.rivers = nrivers

        # Curve Number
        if self.cn is not None:
            ncn = self.cn.copy().rio.clip(polygon.geometry)
            ncn = ncn.where(ncn != -9999)
            ncn = ncn.reindex({'y': self.cn.y, 'x': self.cn.x})
            nwshed.cn = ncn

        nwshed.compute_params(**kwargs)
        return nwshed

    def update_snowlimit(self, snowlimit: Union[int, float],
                         polygonize_kwargs: dict = {},
                         **kwargs: Any) -> Type['RiverBasin']:
        """
        Updates the RiverBasin object to represent only the pluvial (rain-fed) 
        portion of the watershed below a given snowlimit elevation.

        This method clips the basin to areas below the specified snowlimit 
        elevation threshold. The resulting watershed represents only the
        portion of the basin that receives precipitation as rainfall rather
        than snow. All watershed properties (area, rivers, DEM, etc.) are
        updated accordingly.

        Args:
            snowlimit (int|float): Elevation threshold in same units as DEM that 
                defines the rain/snow transition zone
                object with the new clipped one. Defaults to False which leads
                to update only the parameter table and basin masks.
            polygonize_kwargs (dict, optional): Additional keyword arguments 
                passed to the polygonize function. Defaults to {}.
            **kwargs: Additional keyword arguments passed to the compute_params
                method

        Raises:
            TypeError: If snowlimit argument is not numeric

        Returns:
            RiverBasin: A new RiverBasin object containing only the pluvial
                portion of the original watershed below the snowlimit
        """
        if not isinstance(snowlimit, (int, float)):
            raise TypeError("snowlimit must be numeric")
        min_elev = self.dem.elevation.min().item()
        max_elev = self.dem.elevation.max().item()
        if snowlimit < min_elev:
            warnings.warn(f"snowlimit: {snowlimit} below hmin: {min_elev}")
            self.params = self.params*0
            self.mask_raster = xr.DataArray(np.full(self.mask_raster.shape,
                                                    False),
                                            dims=self.mask_raster.dims,
                                            coords=self.mask_raster.coords)
            self.mask_vector = gpd.GeoDataFrame()
            return self
        elif snowlimit > max_elev:
            warnings.warn(f"snowlimit: {snowlimit} above hmax: {max_elev}")
            self.compute_params(**kwargs)
            self.mask_vector = self.basin
            self.mask_raster = ~self.dem.elevation.isnull()
            return self
        else:
            if self.params.loc['area'].item() == 0:
                self.compute_params()
            nshp = polygonize(self.dem.elevation <= snowlimit,
                              **polygonize_kwargs)
            nwshed = self.clip(nshp, **kwargs)
            self.params = nwshed.params
            self.mask_raster = nwshed.mask_raster
            self.mask_vector = nwshed.mask_vector
            return self

    def SynthUnitHydro(self, method: str, ChileParams: bool = False,
                       **kwargs: Any) -> Type['RiverBasin']:
        """
        Compute synthetic unit hydrograph for the basin.

        This method creates and computes a synthetic unit hydrograph based
        on basin parameters. For Chilean watersheds, special regional
        parameters can be used if ChileParams = True.

        Args:
            method (str): Type of synthetic unit hydrograph to use.
                Options: 
                    - 'SCS': SCS dimensionless unit hydrograph
                    - 'Gray': Gray's method
                    - 'Linsley': Linsley method
            ChileParams (bool): Whether to use Chile-specific regional
                parameters. Only valid for 'Gray' and 'Linsley' methods.
                Defaults to False.
            **kwargs: Additional arguments passed to the unit hydrograph
                computation method.

        Returns:
            RiverBasin: Updated instance with computed unit hydrograph stored
                in UnitHydro attribute.

        Raises:
            RuntimeError: If using Chilean parameters and basin centroid lies
                outside valid geographical regions.
        """
        if (method == 'Linsley') and ChileParams:
            poly = CHILE_UH_LINSLEYPOLYGONS
            centroid = self.basin.centroid.to_crs('epsg:4326').loc[0]
            mask = centroid.within(poly.geometry)
            if mask.sum() == 0:
                text = f'Basin {self.fid} is outside the geographical limits'
                text = text+f' allowed by the Chilean {method} method.'
                raise RuntimeError(text)
            else:
                DGAChile_LinsleyZone = poly[mask].zone.item()
                self.params.loc['DGAChile_LinsleyZone'] = DGAChile_LinsleyZone
                uh = SUH(method, self.params[self.fid])
                uh = uh.compute(DGAChileParams=True,
                                DGAChileZone=DGAChile_LinsleyZone, **kwargs)
        elif (method == 'Gray') and ChileParams:
            poly = CHILE_UH_GRAYPOLYGONS
            centroid = self.basin.centroid.to_crs('epsg:4326').loc[0]
            mask = centroid.within(poly.geometry)
            if mask.sum() == 0:
                text = f'Basin {self.fid} is outside the geographical limits'
                text = text+f' allowed by the Chilean {method} method.'
                raise RuntimeError(text)
            else:
                uh = SUH(method, self.params[self.fid])
                uh = uh.compute(DGAChileParams=True, **kwargs)
        else:
            uh = SUH(method, self.params[self.fid])
            uh = uh.compute(**kwargs)
        self.UnitHydro = uh
        return self

    def plot(self,
             demvar='elevation',
             legend_kwargs: dict = {'loc': 'upper left'},
             outlet_kwargs: dict = {'ec': 'k', 'color': 'tab:red'},
             basin_kwargs: dict = {'edgecolor': 'k'},
             demimg_kwargs: dict = {'cbar_kwargs': {'shrink': 0.8}},
             mask_kwargs: dict = {'hatches': ['////']},
             demhist_kwargs: dict = {'alpha': 0.5},
             hypsometric_kwargs: dict = {'color': 'darkblue'},
             rivers_kwargs: dict = {'color': 'tab:red'},
             exposure_kwargs: dict = {'ec': 'k', 'width': 0.6},
             kwargs: dict = {'figsize': (12, 5)}) -> matplotlib.axes.Axes:
        """
        Create a comprehensive visualization of watershed characteristics
            including:
            - 2D map view showing DEM, basin boundary, rivers and outlet point
            - Polar plot showing terrain aspect/exposure distribution
            - Hypsometric curve and elevation histogram

        Args:
            legend (bool, optional): Whether to display legend.
                Defaults to True.
            legend_kwargs (dict, optional): Arguments for legend formatting.
                Defaults to {'loc': 'upper left'}.
            outlet_kwargs (dict, optional): Styling for basin outlet point.
                Defaults to {'ec': 'k', 'color': 'tab:red'}.
            basin_kwargs (dict, optional): Styling for basin boundary.
                Defaults to {'edgecolor': 'k'}.
            demimg_kwargs (dict, optional): Arguments for DEM image display.
                Defaults to {'cbar_kwargs': {'shrink': 0.8}}.
            mask_kwargs  (dict, optional): Arguments for mask hatches.
                Defaults to {'hatches': ['////']}.
            demhist_kwargs (dict, optional): Arguments for elevation histogram.
                Defaults to {'alpha': 0.5}.
            hypsometric_kwargs (dict, optional): Styling for hypsometric curve.
                Defaults to {'color': 'darkblue'}.
            rivers_kwargs (dict, optional): Styling for river network.
                Defaults to {'color': 'tab:red'}.
            exposure_kwargs (dict, optional): Styling for polar exposure plot
                Defaults to {'ec':'k', 'width':0.5}
            kwargs (dict, optional): Additional figure parameters.
                Defaults to {'figsize': (12, 5)}.

        Returns:
            (tuple): Matplotlib figure and axes objects
                (fig, (ax0, ax1, ax2, ax3))
                - ax0: Map view axis
                - ax1: Aspect distribution polar axis  
                - ax2: Hypsometric curve axis
                - ax3: Elevation histogram axis
        """
        # Create figure and axes
        fig = plt.figure(**kwargs)
        ax0 = fig.add_subplot(121)
        ax1 = fig.add_subplot(222, projection='polar')
        ax2 = fig.add_subplot(224)
        ax3 = ax2.twinx()

        # Plot basin and rivers
        try:
            self.basin.boundary.plot(ax=ax0, zorder=2, **basin_kwargs)
            ax0.scatter(self.basin['outlet_x'], self.basin['outlet_y'],
                        label='Outlet', zorder=3, **outlet_kwargs)
        except Exception as e:
            warnings.warn(str(e))

        if len(self.rivers_main) > 0:
            self.rivers_main.plot(ax=ax0, label='Main River', zorder=2,
                                  **rivers_kwargs)

        # Plot dem data
        try:
            self.dem[demvar].plot.imshow(ax=ax0, zorder=0, **demimg_kwargs)
            if len(self.hypsometric_curve) == 0:
                self.get_hypsometric_curve()
            hypso = self.hypsometric_curve
            hypso.plot(ax=ax2, zorder=1, label='Hypsometry',
                       **hypsometric_kwargs)
            ax3.plot(hypso.index, hypso.diff(), zorder=0, **demhist_kwargs)
        except Exception as e:
            warnings.warn(str(e))

        # Plot snow area mask
        try:
            mask = self.mask_raster
            nanmask = self.dem.elevation.isnull()
            if (~nanmask).sum().item() != mask.sum().item():
                mask.where(~nanmask).where(~mask).plot.contourf(
                    ax=ax0, zorder=1, colors=None, alpha=0, add_colorbar=False,
                    **mask_kwargs)
                ax0.plot([], [], label='Snowy Area', color='k')
        except Exception as e:
            warnings.warn(str(e))

        # Plot basin exposition
        if len(self.params.index) > 1:
            exp = self.exposure_distribution
            exp.index = exp.index.map(lambda x: x.split('_')[0])
            exp = exp.loc[['N', 'NE', 'E', 'SE', 'S', 'SW', 'W', 'NW']]
            exp = pd.concat([exp.iloc[:, 0], exp.iloc[:, 0][:'N']])
            ax1.bar(np.deg2rad([0, 45, 90, 135, 180, 225, 270, 315, 0]), exp,
                    **exposure_kwargs)
            ax1.set_xticks(ax1.get_xticks())
            ax1.set_xticklabels(exp.index.values[:-1])
            ax1.set_ylim(0, exp.max()*1.1)

        # Aesthetics
        try:
            for axis in [ax0, ax1, ax2, ax3]:
                axis.set_title('')
                if axis in [ax0, ax2]:
                    axis.legend(**legend_kwargs)
            bounds = self.basin.minimum_bounding_circle().bounds
            ax0.set_xlim(bounds.minx.item(), bounds.maxx.item())
            ax0.set_ylim(bounds.miny.item(), bounds.maxy.item())
            ax1.set_theta_zero_location("N")
            ax1.set_theta_direction(-1)
            ax1.set_xticks(ax1.get_xticks())
            ax1.set_yticklabels([])
            ax1.grid(True, ls=":")

            ax2.grid(True, ls=":")
            ax2.set_ylim(0, 1)
            ax3.set_ylim(0, ax3.get_ylim()[-1])
            ax2.set_xlabel('(m)')

        except Exception as e:
            warnings.warn(str(e))
        return fig, (ax0, ax1, ax2, ax3)


# class RiverReach(object):
#     def tests(self, rivers, dem):
#         """
#         Args:
#             rivers (GeoDataFrame): River network lines
#             dem (xarray.DataArray): Digital elevation model raster
#         Raises:
#             RuntimeError: If any dataset isnt in a projected (UTM) crs.
#         """
#         prj_error = '{} must be in a projected (UTM) crs !'
#         if not rivers.crs.is_projected:
#             error = prj_error.format('Rivers geometry')
#             raise RuntimeError(error)
#         if not dem.rio.crs.is_projected:
#             error = prj_error.format('DEM raster')
#             raise RuntimeError(error)

#     def __init__(self, fid, dem, rivers):
#         """
#         River reach/channel class constructor

#         Args:
#             fid (str): Basin identifier
#             rivers (GeoDataFrame): River network segments
#             dem (xarray.DataArray): Digital elevation model

#         Raises:
#             RuntimeError: If any of the given spatial data isnt in a projected
#                 cartographic projection.
#         """
#         self.tests(rivers, dem)
#         # ID
#         self.fid = fid

#         # Vectors
#         self.rivers = rivers
#         self.rivers_main = pd.Series([])

#         # Terrain
#         self.dem = dem.rio.write_nodata(-9999).squeeze()
#         self.dem = self.dem.to_dataset(name='elevation')
#         self.dem.encoding = dem.encoding


# class Reservoir(object):
#     def __init__(self):
#         pass
